#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import PySimpleGUI as sg
from PySimpleGUI import COLOR_SYSTEM_DEFAULT
def move_window_onto_screen(window: sg.Window):
    # move back on screen if off screen for some reason
    s_w, s_h = window.get_screen_size()
    win_w, win_h = window.size
    loc_x, loc_y = window.current_location()
    if loc_x + win_w >= s_w:
        loc_x = s_w - win_w - 20
    elif loc_x < 0:
        loc_x = 0
    if loc_y + win_h >= s_h:
        loc_y = s_h - win_h - 20
    elif loc_y < 0:
        loc_y = 0
    window.move(loc_x, loc_y)

def set_pep_theme(sg_lib):
    name, theme = "PEP_THEME", {"BACKGROUND": "#FAFAFA", "TEXT": COLOR_SYSTEM_DEFAULT, "INPUT": COLOR_SYSTEM_DEFAULT,
                             "TEXT_INPUT": COLOR_SYSTEM_DEFAULT, "SCROLL": COLOR_SYSTEM_DEFAULT, "BUTTON": COLOR_SYSTEM_DEFAULT,
                             "PROGRESS": COLOR_SYSTEM_DEFAULT, "BORDER": 1, "SLIDER_DEPTH": 1, "PROGRESS_DEPTH": 0, }
    sg_lib.theme_add_new(name, theme)
    sg_lib.theme(name)


#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import abc
import atexit
import os
import shutil
import subprocess
import threading
import time
from datetime import datetime
from time import sleep
from typing import List, Optional, IO

try:
    from Queue import Queue, Empty
except ImportError:
    from queue import Queue, Empty  # python 3.x

from pep_tk.core.job import JobState, JobMeta, TaskStatus, TaskKey
from pep_tk.core.kwiver.pipeline_compiler import compile_output_filenames
from pep_tk.core.kwiver.runner import KwiverRunner


class SchedulerEventManager(metaclass=abc.ABCMeta):
    def __init__(self):
        self.task_start_time = {}
        self.task_end_time = {}
        self.task_status = {}
        self.task_messages = {}
        self.task_count = {}
        self.task_max_count = {}
        self.initialized_tasks = []
        self.task_output_files = {}

    def initialize_task(self, task_key: TaskKey, count: int, max_count: int, status: TaskStatus,
                        task_outputs: Optional[List[str]] = None):
        self.task_count[task_key] = count
        self.task_max_count[task_key] = max_count
        self.task_status[task_key] = status
        self.initialized_tasks.append(task_key)
        self._initialize_task(task_key, count, max_count, status)

        if task_outputs:
            self.update_task_output_files(task_key, task_outputs)

    def start_task(self, task_key: TaskKey):
        self.task_status[task_key] = TaskStatus.RUNNING
        self.task_start_time[task_key] = time.time()
        return self._start_task(task_key)

    def end_task(self, task_key: TaskKey, status: TaskStatus):
        self.task_end_time[task_key] = time.time()
        self.task_status[task_key] = status
        return self._end_task(task_key, status)

    def check_cancelled(self, task_key: TaskKey) -> bool:
        return self._check_cancelled(task_key)

    def update_task_progress(self, task_key: TaskKey, current_count: int):
        self.task_count[task_key] = current_count
        return self._update_task_progress(task_key, current_count, self.task_max_count[task_key])

    def update_task_stdout(self, task_key: TaskKey, line: str):
        self._update_task_stdout(task_key, line)

    def update_task_stderr(self, task_key: TaskKey, line: str):
        self._update_task_stdout(task_key, line)

    def update_task_output_files(self, task_key: TaskKey, output_files: List[str]):
        self.task_output_files[task_key] = output_files
        return self._update_task_output_files(task_key, output_files)

    def elapsed_time(self, task_key: TaskKey) -> float:
        if task_key not in self.task_start_time:
            return 0.
        if task_key not in self.task_end_time:
            return time.time() - self.task_start_time[task_key]
        return self.task_end_time[task_key] - self.task_start_time[task_key]

    @abc.abstractmethod
    def _initialize_task(self, task_key: TaskKey, count: int, max_count: int, status: TaskStatus):
        pass

    @abc.abstractmethod
    def _start_task(self, task_key: TaskKey):
        pass

    @abc.abstractmethod
    def _end_task(self, task_key: TaskKey, status: TaskStatus):
        pass

    @abc.abstractmethod
    def _update_task_progress(self, task_key: TaskKey, current_count: int, max_count: int):
        pass

    @abc.abstractmethod
    def _update_task_stdout(self, task_key: TaskKey, line: str):
        pass

    @abc.abstractmethod
    def _update_task_stderr(self, task_key: TaskKey, line: str):
        pass

    @abc.abstractmethod
    def _check_cancelled(self, task_key: TaskKey) -> bool:
        pass

    @abc.abstractmethod
    def _update_task_output_files(self, task_key: TaskKey, output_files: List[str]):
        pass


# Scheduler helpers
def poll_image_list(fp: str):
    if not os.path.exists(fp):
        return 0
    with open(fp, 'r') as f:
        count = sum(1 for _ in f)
    return count


def monitor_outputs(stop_event: threading.Event, task_key: TaskKey, manager: SchedulerEventManager,
                    output_file: str, poll_freq: int):
    while not stop_event.wait(poll_freq):
        try:
            count = poll_image_list(output_file)
            manager.update_task_progress(task_key, count)
        except Exception as e:
            # should not have an issue but this is just to ensure program doesn't crash for user
            # pretty sure the concurrency stuff here is solid, but hard to be certain
            print(e)


def enqueue_output(out, queue, evt: threading.Event, logfile: IO):
    # don't want it to stop on an emty byte(b'') because we need to detect
    # if an empty byte has come through and we can't do it asynchronously
    for line in iter(out.readline, b'foobar'):
        if evt.is_set():
            logfile.close()
            out.close()
            break
        try:
            logfile.write(line)
            queue.put(line)
        except Exception as e:
            # should not have an issue but this is just to ensure program doesn't crash for user
            # pretty sure the concurrency stuff here is solid, but hard to be certain
            print(e)


def move_output_files(output_fps, destination_dir):
    new_files = []
    for current_loc in output_fps:
        if not os.path.isfile(current_loc):
            # handle case where output file doesn't exist (errored before pipeline started)
            continue
        new_loc = os.path.join(destination_dir, os.path.basename(current_loc))
        shutil.move(current_loc, new_loc)
        new_files.append(new_loc)
    return new_files


def exit_cleanup(fds, files_to_move, dir_to_move):
    for f in fds:
        try:
            f.close()
        except:
            print(f'Warning: Unable to close {f.name}')

    move_output_files(files_to_move, dir_to_move)


def kill_process(process):
    if os.name == 'nt':
        subprocess.call(['taskkill', '/F', '/T', '/PID', str(process.pid)])
    else:
        process.kill()


class Scheduler:
    def __init__(self,
                 job_state: JobState,
                 job_meta: JobMeta,
                 manager: SchedulerEventManager,
                 kwiver_setup_path: str,
                 progress_poll_freq: int = 1,
                 kill_event: threading.Event() = None):
        """
        Initialize a Scheduler for proccessing task queue synchronously.

        :param job_state: the job state
        :param job_meta: the job metadata
        :param manager: manager for dispatching events, updating progress, etc..
        :param kwiver_setup_path: setup_viame.sh/bat path
        :param progress_poll_freq: frequency to poll progress (reads output file and counts progress)
        :param kill_event: threading.Event to send the scheduler if the GUI thread is exited or the program is killed
        which will trigger the scheduler to cleanup and exit cleanly
        """
        self.job_state = job_state
        self.job_meta = job_meta
        self.manager = manager
        self.kwiver_setup_path = kwiver_setup_path
        self.progress_poll_freq = progress_poll_freq
        self.kill_event: threading.Event() = kill_event

    def run(self):
        print('Scheduler Started (pid: %d)' % os.getpid())
        # if resuming mark already completed tasks as completed
        for task_key in self.job_state.tasks(status=TaskStatus.SUCCESS):
            pipeline_fp, dataset, outputs = self.job_meta.get(task_key)
            max_image_count = max(dataset.thermal_image_count, dataset.color_image_count)

            # read output log if exists
            stdout_log_fp = os.path.join(self.job_meta.logs_dir,
                                         f'kwiver-output-{task_key.replace(":", "_")}.log')
            if os.path.isfile(stdout_log_fp):
                with open(stdout_log_fp, 'r') as f:
                    log = f.read()
                self.manager.update_task_stdout(task_key,
                                                'Task already complete.  Log file found: %s\n' % stdout_log_fp)
                self.manager.update_task_stdout(task_key, log)

            # initialize the task
            task_outputs = self.job_state.get_task_outputs(task_key)
            self.manager.initialize_task(task_key, max_image_count, max_image_count, TaskStatus.SUCCESS, task_outputs)

        for task_key in self.job_state.tasks():
            status = self.job_state.get_status(task_key)
            if status != TaskStatus.SUCCESS:
                pipeline_fp, dataset, outputs = self.job_meta.get(task_key)
                max_image_count = max(dataset.thermal_image_count, dataset.color_image_count)
                self.manager.initialize_task(task_key, 0, max_image_count, TaskStatus.INITIALIZED)

        while not self.job_state.is_job_complete():
            current_task_key = self.job_state.current_task()
            pipeline_fp, dataset, outputs = self.job_meta.get(current_task_key)

            # Create the environment variables needed for running
            #  - output ports (image list and viame detection csv file names)
            #  - the kwiver environment required for running kwiver runner
            stdout_enqueue_thread = datetime.now()
            csv_ports_raw = outputs.get_det_csv_env_ports()
            image_list_raw = outputs.get_image_list_env_ports()

            pipeline_output_csv_env = compile_output_filenames(csv_ports_raw, path=self.job_meta.pending_outputs_dir,
                                                               t=stdout_enqueue_thread)
            pipeline_output_image_list_env = compile_output_filenames(image_list_raw,
                                                                      path=self.job_meta.pending_outputs_dir,
                                                                      t=stdout_enqueue_thread)

            env = {**pipeline_output_csv_env, **pipeline_output_image_list_env}

            # Setup error log
            stdout_log_fp = os.path.join(self.job_meta.logs_dir,
                                         f'kwiver-output-{current_task_key.replace(":", "_")}.log')

            output_log = open(stdout_log_fp, 'w+b')

            atexit.register(exit_cleanup, fds=[output_log],
                            files_to_move=list(pipeline_output_csv_env.values()) + list(
                                pipeline_output_image_list_env.values()),
                            dir_to_move=self.job_meta.error_outputs_dir)

            # Update Task Started
            self.manager.start_task(current_task_key)
            self.job_state.set_task_status(current_task_key, TaskStatus.RUNNING)

            # create the progress polling thread and start it
            image_list_monitor = list(pipeline_output_image_list_env.values())[0]  # Image list to monitor
            prog_stop_evt = threading.Event()
            thread_args = (prog_stop_evt, current_task_key, self.manager, image_list_monitor, self.progress_poll_freq)
            progress_thread = threading.Thread(target=monitor_outputs,
                                               args=thread_args,
                                               daemon=True)
            progress_thread.start()

            # Create the kwiver runner and run it
            kwr = KwiverRunner(pipeline_fp,
                               cwd=self.job_meta.root_dir,
                               env=env,
                               kwiver_setup_path=self.kwiver_setup_path)

            process = kwr.run(stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            print('Kwiver Runner Started (pid: %d)' % process.pid)

            if process.stdout is None:
                raise RuntimeError("Stdout must not be none")

            # Stdout Thread
            kwiver_output_queue = Queue()
            stdout_enqueue_thread = threading.Thread(target=enqueue_output, args=(
            process.stdout, kwiver_output_queue, prog_stop_evt, output_log))
            stdout_enqueue_thread.daemon = True  # thread dies with the program
            stdout_enqueue_thread.start()

            cancelled = False
            while not cancelled:  # read line without blocking
                if self.kill_event:
                    if self.kill_event.is_set():
                        # Kill all incomplete tasks
                        self._kill_all_tasks(process, prog_stop_evt)
                        exit_cleanup(fds=[output_log],
                                     files_to_move=list(pipeline_output_csv_env.values()) + list(
                                         pipeline_output_image_list_env.values()),
                                     dir_to_move=self.job_meta.error_outputs_dir)
                        return
                try:
                    line = kwiver_output_queue.get(timeout=.5)
                    if line == b'':
                        break  # job is complete if empty byte received
                    else:
                        self.manager.update_task_stdout(current_task_key, line.decode("utf-8"))
                except Empty:
                    pass

                # check if user cancelled task, if cancelled kill kwiver process and stop output reading loop
                cancelled = self.manager.check_cancelled(current_task_key)

            # Wait for exit up to 30 seconds after kill
            code = process.wait(30)

            # stop polling for progress and stop polling for stdout
            prog_stop_evt.set()

            outputs_to_move = list(pipeline_output_csv_env.values()) + list(pipeline_output_image_list_env.values())

            # if user cancells task
            if cancelled:
                kill_process(process)
                print(f'Cancelled {current_task_key}')

                count = poll_image_list(image_list_monitor)
                self.manager.update_task_progress(current_task_key, count)
                self.job_state.set_task_status(current_task_key, TaskStatus.CANCELLED)
                self.manager.end_task(current_task_key, TaskStatus.CANCELLED)

                # Move outputs to error folder, attempt until process releases lock on files
                moved = False
                attempts = 0
                while not moved:
                    if attempts > 30:
                        # try for 30 seconds
                        break
                    try:
                        move_output_files(outputs_to_move, self.job_meta.error_outputs_dir)
                        moved = True
                    except PermissionError:
                        sleep(1)
                        attempts += 1

                continue

            if code > 0:  # ERROR
                # Update Task Ended with error
                count = poll_image_list(image_list_monitor)
                self.manager.update_task_progress(current_task_key, count)
                self.job_state.set_task_status(current_task_key, TaskStatus.ERROR)
                self.manager.end_task(current_task_key, TaskStatus.ERROR)

                # Move output files to error dir
                outputs_to_move = list(pipeline_output_csv_env.values()) + list(pipeline_output_image_list_env.values())
                move_output_files(outputs_to_move, self.job_meta.error_outputs_dir)
            else:  # SUCCESS
                # Update Task final count in GUI, has to be done before moving file
                count = poll_image_list(image_list_monitor)
                self.manager.update_task_progress(current_task_key, count)

                # Move outputs to completed folder
                outputs_new_loc = move_output_files(outputs_to_move, self.job_meta.completed_outputs_dir)

                # Update Task State with success and output files
                self.job_state.set_task_outputs(current_task_key, outputs_new_loc)
                self.job_state.set_task_status(current_task_key, TaskStatus.SUCCESS)

                # Update GUI with success
                self.manager.end_task(current_task_key, TaskStatus.SUCCESS)
                self.manager.update_task_output_files(current_task_key, outputs_new_loc)


    def _kill_all_tasks(self, process: subprocess.Popen, prog_stop_evt: threading.Event):
        for task_to_end in self.job_state.tasks():
            if self.job_state.is_task_complete(task_to_end):
                continue  # do not modify state of complete tasks

            # set all tasks statuses to cancelled
            self.job_state.set_task_status(task_to_end, TaskStatus.ERROR)
            self.manager.end_task(task_to_end, TaskStatus.ERROR)
        prog_stop_evt.set()
        process.kill()
        process.wait(timeout=30)


#     This file is part of the PEP GUI detection pipeline batch running tool
#     Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.
def main():
    from pep_tk.core.parser import load_dataset_manifest, DatasetManifestError, EmptyParser
    from pep_tk.psg.settings import UserProperties
    from pep_tk.psg.windows import launch_gui, popup_error
    from pep_tk.core.configuration import PipelineManifest
    success = False
    while not success:
        # launch_gui returns False if needs to be refreshed, returns true if program exits
        pm = PipelineManifest()
        try:
            p = UserProperties()
            dm = load_dataset_manifest(p.data_manifest_filepath)
        except DatasetManifestError as e:
            # Handeled dataset manifest exception on startup - pass an empty dataset manifest
            if hasattr(e, 'message'):
                msg = f'Error Type: {e.__class__.__name__}\nMessage:\n{e.message}'
            else:
                msg = str(e)
            dm = EmptyParser(error_message=msg)
        except Exception as e:
            # Unhandled exception on startup - show error and end application
            if hasattr(e, 'message'):
                msg = f'Unhandled Error Contact Yuval!\nError Type:\n{e.__class__.__name__}\nMessage:\n{e.message}'
            else:
                msg = str(e)

            popup_error(msg)
            return

        success = launch_gui(pm, dm)


if __name__ == "__main__":
    # PLUGIN_PATH = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
    # plugin_python_lib = os.path.join(PLUGIN_PATH, 'lib', 'python3.6', 'site-packages')
    # sys.path.insert(0, plugin_python_lib)
    # print('Added %s to PYTHONPATH' % plugin_python_lib)

    main()



#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import shutil
from enum import Enum
from typing import List, Tuple, Optional

from pep_tk.core.utilities import jsonfile
from pep_tk.core.parser import VIAMEDataset
from pep_tk.core.configuration import PipelineConfig
from pep_tk.core.configuration.configurations import PipelineOutputOptionGroup
from pep_tk.core.kwiver.pipeline_compiler import compile_pipeline


class JobInitException(Exception):
    pass


# get job directories / files using the job root dir
meta_dir = lambda root_dir: os.path.join(root_dir, 'meta')
pipelines_dir = lambda root_dir: os.path.join(root_dir, 'pipelines')
logs_dir = lambda root_dir: os.path.join(root_dir, 'logs')
completed_outputs_dir = lambda root_dir: os.path.join(root_dir, 'outputs_success')
error_outputs_dir = lambda root_dir: os.path.join(root_dir, 'outputs_error')
pending_outputs_dir = lambda root_dir: os.path.join(root_dir, 'outputs_pending')

job_state_json_fp = lambda root_dir: os.path.join(meta_dir(root_dir), 'job_state.json')
pipeline_meta_json_fp = lambda root_dir: os.path.join(meta_dir(root_dir), 'pipelines_meta.json')
pipeline_manifest_local = lambda root_dir: os.path.join(meta_dir(root_dir), 'pipelines_manifest.yaml')
datasets_meta_json_fp = lambda root_dir: os.path.join(meta_dir(root_dir), 'datasets_meta.json')


class JobMeta:
    def __init__(self, root_dir):
        self.root_dir = root_dir
        self.logs_dir = logs_dir(root_dir)
        self.completed_outputs_dir = completed_outputs_dir(root_dir)
        self.error_outputs_dir = error_outputs_dir(root_dir)
        self.pending_outputs_dir = pending_outputs_dir(root_dir)

        self.pipe_meta_fp = pipeline_meta_json_fp(root_dir)
        self.dataset_meta_fp = datasets_meta_json_fp(root_dir)
        self.compiled_pipelines_dir = pipelines_dir(root_dir)  # where the compiled pipelines go
        self._was_existing = os.path.isfile(self.dataset_meta_fp) and os.path.isfile(self.pipe_meta_fp)

        dump_kwargs = dict(ensure_ascii=False, indent="\t", sort_keys=True)
        self._pipe_store = jsonfile.jsonfile(self.pipe_meta_fp, default_data={}, autosave=True, dump_kwargs=dump_kwargs)
        self._ds_store = jsonfile.jsonfile(self.dataset_meta_fp, default_data={}, autosave=True,
                                           dump_kwargs=dump_kwargs)

    def create_meta(self, pipeline: PipelineConfig, datasets: List[VIAMEDataset]):
        self._pipe_store.data = pipeline.to_dict()
        for idx, dataset in enumerate(datasets):
            compiled_fp = os.path.join(self.compiled_pipelines_dir,
                                       f'{dataset.filename_friendly_name}-{pipeline.name}.pipe')

            # compile output ports first so we can cache output information
            output_config = pipeline.output_group.to_dict()
            for config_name, v in output_config.items():
                output_pattern = v['default'].replace('[DATASET]', dataset.filename_friendly_name)
                output_config[config_name]['_value'] = output_pattern
                output_config[config_name]['_locked'] = True

            # compile everything EXCEPT the new outputs
            env = {**pipeline.get_parameter_env_ports(),
                   **pipeline.get_pipeline_dataset_environment(dataset)}
            compiled_pipe = compile_pipeline(pipeline, env)
            with open(compiled_fp, 'w') as f:
                f.write(compiled_pipe)

            compiled_relpath = os.path.relpath(compiled_fp, self.root_dir)
            self._ds_store.data[dataset.name] = {'compiled_fp': compiled_relpath, 'dataset': dataset.asdict(),
                                                 'output_config': output_config}

    def keys(self):
        return list(self._ds_store.data.keys())

    def get(self, dataset_key) -> Optional[Tuple[str, VIAMEDataset, PipelineOutputOptionGroup]]:
        ds_meta = self._ds_store.data.get(dataset_key)
        if ds_meta is None:
            return None
        # pipeline_fp = os.path.join(self.root_dir, ds_meta['compiled_fp'])
        pipeline_fp = ds_meta['compiled_fp']
        ds_obj = VIAMEDataset(**ds_meta['dataset'].asdict())
        outputs = PipelineOutputOptionGroup(ds_meta)

        return pipeline_fp, ds_obj, outputs


TaskKey = str


class TaskStatus(Enum):
    INITIALIZED = -1
    ERROR = 0
    SUCCESS = 1
    RUNNING = 2
    CANCELLED = 3


class JobState:
    def __init__(self, root_dir, pipeline_keys=None, load_existing=False):
        self.state_fp = job_state_json_fp(root_dir)

        dump_kwargs = dict(ensure_ascii=False, indent="\t", sort_keys=True)
        self._store = jsonfile.jsonfile(self.state_fp, default_data={}, autosave=True, dump_kwargs=dump_kwargs)

        if load_existing:
            if not os.path.isfile(self.state_fp):
                msg = f'Unable to load job. {self.state_fp} does not exist.'
                raise JobInitException(msg)

            self._store = jsonfile.jsonfile(self.state_fp, default_data={}, autosave=True, dump_kwargs=dump_kwargs)
            if not self._store.data.get('initialized', False):
                msg = f'Possibly corrupt job file, please share the following file with Yuval. {self.state_fp}'
                raise JobInitException(msg)
        else:
            # build the job state cache in the provided directory
            if self._store.data.get('initialized', False):
                raise JobInitException('Job Already Exists, this is a check to make sure jobs cant be overriden.')

            if not pipeline_keys or len(pipeline_keys) < 1:
                raise JobInitException('No pipelines provided.')

            # initialize the new job
            self._store.data['tasks'] = sorted(pipeline_keys)
            self._store.data['task_status'] = {task_key: TaskStatus.INITIALIZED.value for task_key in pipeline_keys}
            self._store.data['total_tasks'] = len(pipeline_keys)
            self._store.data['task_outputs'] = {task_key: [] for task_key in pipeline_keys}
            self._store.data['initialized'] = True

        # reset any previous errored tasks to initialized
        for task_key in self._store.data['task_status']:
            if self.get_status(task_key) != TaskStatus.SUCCESS:
                self._store.data['task_status'][task_key] = TaskStatus.INITIALIZED.value

    def get_status(self, task_key: TaskKey):
        return TaskStatus(self._store.data['task_status'][task_key])

    @classmethod
    def load(cls, meta_directory):
        return cls(meta_directory, load_existing=True)

    def current_task(self) -> Optional[TaskKey]:
        for task_key in self.tasks():
            if self.is_task_complete(task_key):
                continue
            return task_key

        return None

    def is_task_complete(self, task_key: TaskKey) -> bool:
        task_status = self.get_status(task_key)
        is_complete = task_status in [TaskStatus.SUCCESS, TaskStatus.ERROR, TaskStatus.CANCELLED]
        return is_complete

    def set_task_status(self, task_key: TaskKey, status: TaskStatus):
        self._store.data['task_status'][task_key] = status.value

    def set_task_outputs(self, task_key: TaskKey, outputs: List[str]):
        self._store.data['task_outputs'][task_key] = outputs

    def get_task_outputs(self, task_key: TaskKey) -> Optional[List[str]]:
        if len(self._store.data['task_outputs'][task_key]) == 0:
            return None
        else:
            return list(self._store.data['task_outputs'][task_key])

    def is_job_complete(self) -> bool:
        return all([self.is_task_complete(task_key) for task_key in self.tasks()])

    def tasks(self, status: TaskStatus = None) -> List[TaskKey]:
        tasks = list(self._store.data['tasks'])
        if status is None:
            return tasks

        ret = []
        for task in tasks:
            if self.get_status(task_key=task) == status:
                ret.append(task)

        return ret

    def completed_tasks(self) -> List[TaskKey]:
        completed = []
        for task_key in self.tasks():
            if self.is_task_complete(task_key):
                completed.append(task_key)
        return completed


def load_job(directory: str) -> Tuple[JobState, JobMeta]:
    job_state = JobState.load(directory)
    job_meta = JobMeta(directory)
    return job_state, job_meta

def job_exists(job_path: str):
    try:
        job_state, job_meta = load_job(job_path)
    except:
        return False
    return True

def create_job(directory, pipeline: PipelineConfig, datasets: List[VIAMEDataset], force=False) -> str:
    if os.path.isdir(directory) or os.path.isfile(directory):
        if force:
            shutil.rmtree(directory, ignore_errors=True)
        else:
            raise Exception('Either directory already exists or is a file an not a directory')

    pipeline_directory = pipelines_dir(directory)
    meta_directory = meta_dir(directory)
    logs_directory = logs_dir(directory)
    error_outputs_directory = error_outputs_dir(directory)
    os.makedirs(directory, exist_ok=False)
    os.makedirs(pipeline_directory, exist_ok=False)
    os.makedirs(meta_directory, exist_ok=False)
    os.makedirs(logs_directory, exist_ok=False)
    os.makedirs(error_outputs_directory, exist_ok=False)
    os.makedirs(completed_outputs_dir(directory), exist_ok=False)
    os.makedirs(pending_outputs_dir(directory), exist_ok=False)
    try:
        # initialize state and meta
        # TODO make interface for initializing job state and meta the same
        job_meta = JobMeta(directory)
        job_meta.create_meta(pipeline=pipeline, datasets=datasets)
        job_state = JobState(directory, job_meta.keys())
    except Exception as e:
        # clean up if failed for some reason
        # important to make sure this is never reached without the initial directory exists check
        # because we don't wan't to delete a pre-existing job directory
        shutil.rmtree(directory)
        raise e
    return directory



#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import base64
import os
import PySimpleGUI as sg
from pep_tk import PLUGIN_PATH
from pep_tk.psg.fonts import Fonts


class SystemSettingsNames:
    viame_directory = 'setup_viame_filepath'
    dataset_manifest_filepath = 'dataset_manifest_filepath'
    window_location = 'window_location'
    recent_jobs_list = 'recent_jobs_list'
    detection_output_location = 'detection_output_location'
    job_directory = 'job_directory'


def image_resource_path(file_path=''):
    return os.path.join(PLUGIN_PATH, 'lib', 'img', file_path)


# filename of icon and '1x' or '2x' for size
def icon_filepath(icon_fn, size='1x'):
    assert (size == '1x' or size == '2x')
    return os.path.normpath(os.path.join(PLUGIN_PATH, 'lib', 'img', 'icons', size, icon_fn))


def system_settings_filepath():
    return os.path.normpath(os.path.join(PLUGIN_PATH, 'peptk_gui_settings.json'))


def get_system_settings():
    return sg.UserSettings(filename=system_settings_filepath())


def get_user_settings():
    return sg.UserSettings(filename='peptk_gui_settings.json')


def get_viame_bash_or_bat_file_path(viame_dir):
    if os.name == 'nt':
        fn = 'setup_viame.bat'
    else:
        fn = 'setup_viame.sh'

    return os.path.normpath(os.path.join(viame_dir, fn))


class UserProperties:
    """ This class represents all of the properties in the settings file that the user has control over, and
     are able to set through the Properties window in the GUI. """
    def __init__(self):
        self.settings = get_user_settings()

    def refresh(self):
        self.settings = get_user_settings()

    @property
    def viame_dir(self):
        return self.settings.get(SystemSettingsNames.viame_directory, None)

    @property
    def data_manifest_filepath(self):
        return self.settings.get(SystemSettingsNames.dataset_manifest_filepath, None)

    @property
    def job_base_dir(self):
        return self.settings.get(SystemSettingsNames.job_directory, None)

    def as_dict(self):
        return {'viame_dir': self.viame_dir,
                'data_manifest_filepath': self.data_manifest_filepath,
                'job_base_dir': self.job_base_dir}


WINDOW_ICON = image_resource_path('icon_80x80.png')
with open(WINDOW_ICON, 'rb') as f:
    icon = base64.b64encode(f.read())
sg.set_options(icon=icon, titlebar_icon=icon, font=Fonts.description)



#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
from dataclasses import dataclass
from typing import List, Optional

from pep_tk.core.job import TaskStatus


@dataclass
class ProgressGUIEventData:
    """ Data message communicated between the Scheduler thread and the main GUI thread """
    progress_count: int = None  # number of items processed already in task
    max_count: int = None  # number of items being processed in task
    elapsed_time: float = None  # time elapsed so far
    task_status: TaskStatus = None  # current status of the task
    output_files: List[str] = None  # output files from the task (image lists, detections)
    output_log: str = None
    completed_on_load: bool = False # if task was already completed when initialized/loaded

    @property
    def time_per_count(self) -> float:  # average time taken to process each item
        if self.progress_count == None:
            return 0
        if self.progress_count == 0:
            return 0
        return self.elapsed_time / self.progress_count

    @property
    def estimated_time_remaining(self) -> Optional[float]:
        if self.progress_count == None or self.max_count == None:
            return None
        return self.time_per_count * (self.max_count - self.progress_count)


#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

# from dataclasses import dataclass
#
#
from dataclasses import dataclass

class ImageList:
    def __init__(self, list_file):
        self.files = []
        base_dir = os.path.dirname(list_file)
        with open(list_file) as f:
            data = f.readlines()
            for d in data:
                fn = os.path.normpath(d.strip())
                if not os.path.isabs(fn):
                    fn = os.path.normpath(os.path.join(base_dir, fn))
                self.files.append(fn)
        self.files = sorted(self.files)
        self.current = -1
        self.total = len(self.files)

    def __iter__(self):
        return self

    def __next__(self): # Python 2: def next(self)
        self.current += 1
        if self.current < self.total:
            return self.files[self.current]
        raise StopIteration

    def __getitem__(self, index: int) -> str:
        return self.files[index]

    def __len__(self):
        return len(self.files)

@dataclass
class VIAMEDataset:
    name: str
    thermal_image_list: str
    color_image_list: str
    transformation_file: str
    _thermal_images: ImageList = None
    _color_images: ImageList = None

    def __getitem__(self, item):
        return getattr(self, item)

    def get(self, item: str, default=None):
        return getattr(self, item, default)

    def __contains__(self, item):
        return hasattr(self, item)

    @property
    def thermal_images(self):
        if not self._thermal_images:
            if self.thermal_image_list:
                self._thermal_images = ImageList(self.thermal_image_list)
        return self._thermal_images if self._thermal_images else None

    @property
    def color_images(self):
        if not self._color_images:
            if self.color_image_list:
                self._color_images = ImageList(self.color_image_list)
        return self._color_images if self._color_images else None

    @property
    def thermal_image_count(self):
        return 0 if not self.thermal_images else len(self.thermal_images)

    @property
    def color_image_count(self):
        return 0 if self.color_images is None else len(self.color_images)

    @property
    def filename_friendly_name(self):
        def safe_char(c):
            if c.isalnum() or c == '-':
                return c
            else:
                return "_"

        return "".join(safe_char(c) for c in self.name).rstrip("_")

    def asdict(self):
        return {'color_image_list': self.color_image_list,
                'thermal_image_list': self.thermal_image_list,
                'transformation_file': self.transformation_file,
                'name': self.name}

#
import os
from abc import ABC, abstractmethod
from typing import List, Optional


class DatasetManifestError(Exception):
    def __init__(self, message, *args: object) -> None:
        self.message = message
        super().__init__(*args)


class DatasetFileNotFound(DatasetManifestError):
    pass

class DuplicateDatasetName(DatasetManifestError):
    pass

class ImageListMissingImage(DatasetManifestError):
    pass

class ParserNotFoundException(DatasetManifestError):
    pass

class NoImageListException(DatasetManifestError):
    pass

class MisingDatasetNameException(DatasetManifestError):
    pass

def path_to_absolute(cwd, path):
    if os.path.isabs(path):
        return os.path.normpath(path)
    else:
        return os.path.normpath(os.path.join(cwd, path))


class ManifestParser(ABC):
    att_thermal_image_list = 'thermal_image_list'
    att_color_image_list = 'color_image_list'
    att_transform = 'transformation_file'

    @abstractmethod
    def list_dataset_keys_txt(self, txt: str) -> List[str]:
        pass

    @abstractmethod
    def list_dataset_keys(self) -> List[str]:
        pass

    @abstractmethod
    def get_dataset(self, name: str) -> Optional[VIAMEDataset]:
        pass

    def __getitem__(self, item) -> Optional[VIAMEDataset]:
        return self.get_dataset(item)



#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import subprocess
from typing import Dict, List


def get_kwiver_runner_command(kwiver_setup_path=None, debug=False) -> List:
    """
    Get Command for Kwiver Runner

    :param kwiver_setup_path: path to kwiver_setup.sh or on windows kwiver_setup.bat
    :param debug: run kwiver with gdb --debug
    :return: List command
    """
    if os.name == 'nt':
        if kwiver_setup_path:
            return [f'"{kwiver_setup_path}"', '&&', 'kwiver.exe', 'runner']
        else:
            return ['kwiver.exe', 'runner']
    else:
        if debug:
            args = ['gdb', '--args', 'kwiver', 'runner']
        else:
            args = ['kwiver', 'runner']

        if kwiver_setup_path:
            args = ['source', kwiver_setup_path, '&&', 'printenv', '&&'] + args
        return args


def execute_command(cmd: str, env: Dict, cwd, stdout=None, stderr=None) -> subprocess.Popen:
    """
    Executes a command using subprocess.Popen on windows or linux.

    :param cmd: command to run
    :param env: environment
    :param cwd: working directory
    :param stdout: stdout file/stream
    :param stderr: stderr file/stream
    :return: subprocess.Popen object
    """
    if os.name == 'nt':
        env = {**env, **os.environ}
        return subprocess.Popen(cmd, cwd=cwd, stdout=stdout, stderr=subprocess.STDOUT, env=env)
    else:
        env = {**env, **os.environ}
        return subprocess.Popen(cmd, cwd=cwd, stdout=stdout, stderr=stderr, env=env, shell=True, executable='/bin/bash')


class KwiverRunner:
    """
        Runner for running non-embedded pipelines that have no input output ports.
        Uses subprocess to call kwiver_runner
    """

    def __init__(self, pipeline_fp: str,
                 cwd: str,
                 env: Dict = None,
                 pipe_args: Dict = None,
                 kwiver_setup_path: str = None):
        """

        :param pipeline_fp: Filepath of the .pipe file being run.
        :param cwd: the working directory in which to run the pipeline.
        :param env: environment variables to set when running the pipeline.
        :param pipe_args: arguments to pass to kwiver runner in the '-s process:param=value' format.
        :param kwiver_setup_path: Path to the 'setup_viame.sh' or on windows 'setup_viame.bat'.
        """
        self.pipeline_fp = pipeline_fp
        self.cwd = cwd
        self.kwiver_setup_path = kwiver_setup_path
        self.env = env or {}
        self.pipe_args = pipe_args or {}

    def get_environment_str(self) -> str:
        """
        Creates the environment string that goes before the command to set the environment of the process
        :return: str command to set environment variables on windows or linux
        """
        if os.name == 'nt':
            env_str = ""
            for k, v in self.env.items():
                env_str += 'SET %s=%s & ' % (k, v) + env_str
            env_str = env_str[:-2]
            return env_str
        else:
            env_str = ""
            for k, v in self.env.items():
                env_str += '%s=%s ' % (k, v) + env_str
            return env_str

    def get_arguments_str(self) -> str:
        """
        Get args string

        :return: string of arguments to pass to kwiver runner
        """
        args_str = ""
        for k, v in self.pipe_args.items():
            args_str += '-s %s=%s ' % (k, v)
        return args_str

    def run(self, stdout=None, stderr=None) -> subprocess.Popen:
        """
        Runs the pipeline with the provided configuration and returns the subprocess.Popen object.

        :param stdout: Optional stream for process to write stdout to
        :param stderr: Optional stream for process to write stderr to
        :return: The subprocess.Popen object
        """
        cmd = get_kwiver_runner_command(kwiver_setup_path=self.kwiver_setup_path) + [self.pipeline_fp]
        cmd = ' '.join(cmd)

        # Add kwiver runner pipeline arguments
        pipe_args_str = self.get_arguments_str()
        if len(pipe_args_str) > 0:
            cmd += " " + pipe_args_str

        # Useful to print the command as environment variables being set in Popen we need to be able to reproduce
        # within the command for debugging purposes.
        print(cmd.replace('&&', '&& ' + self.get_environment_str()) + cmd + ' ' + pipe_args_str)
        proc = execute_command(cmd, self.env, self.cwd, stdout=stdout, stderr=stderr)
        return proc



#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os

import re

class ConfigType:
    def __init__(self, var_type):
        self.var_type = var_type

    def validate(self, text):  raise NotImplementedError('ConfigType: Must implement validate')

    def description(self): raise NotImplementedError('ConfigType: Must implement description')

class StringType(ConfigType):
    def __init__(self):
        ConfigType.__init__(self, var_type=str)

    def validate(self, text):
        return True, text

    def description(self): return 'text'

class ConfigIntType(ConfigType):
    def __init__(self, min, max):
        ConfigType.__init__(self, var_type=int)
        self.min = min
        self.max = max

    def validate(self, text):
        text = str(text)
        if not text.isdigit():
            return False, text
        try:
            text = self.var_type(text)
        except:
            return False, text
        if not self.min is None and text < self.min:
            return False, text
        if not self.max is None and text > self.max:
            return False, text
        return True, text

    def description(self):
        if self.min is not None and self.max is not None: return 'integer between %d and %d' % (self.min, self.max)
        if self.min is None and self.max is not None: return 'integer less than %d' % self.max
        if self.max is None and self.min is not None: return 'integer greater than %d' % self.min
        return 'integer'

class ConfigFloatType(ConfigType):
    def __init__(self, min, max):
        ConfigType.__init__(self, var_type=float)
        self.min = min
        self.max = max

    def validate(self, text):
        try:
            text = self.var_type(text)
        except:
            return False, text
        if not self.min is None and text < self.min:
            return False, text
        if not self.max is None and text > self.max:
            return False, text
        return True, text

    def description(self):
        if self.min is not None and self.max is not None: return 'decimal between %.1f and %.1f' % (self.min, self.max)
        if self.min is None and self.max is not None: return 'decimal less than %.1f' % self.max
        if self.max is None and self.min is not None: return 'decimal greater than %.1f' % self.min
        return 'decimal'

class ConfigOutputImageListType(StringType):
    def __init__(self):
        StringType.__init__(self)

    def validate(self, text):
        fp_pattern, ext = os.path.splitext(text)
        return ext=='.txt', fp_pattern

    def description(self): return 'file pattern'

class ConfigOutputDetectionsType(StringType):
    def __init__(self):
        StringType.__init__(self)

    def validate(self, text):
        fp_pattern, ext = os.path.splitext(text)
        return ext=='.csv', fp_pattern

    def description(self): return 'file pattern'


def parse_type(type_str: str):
    '''
    Parse a type
    int[]
    str
    float
    '''
    int_exp = r'^int(?:\[(\d+)\,(\d+)?\])?$'
    float_exp = r'^float(?:\[([+-]?(?:[0-9]*[.])?[0-9]+)\,([+-]?(?:[0-9]*[.])?[0-9]+)?\])?$'
    out_imagelist_exp = '^output_image_list$'
    out_detections_exp = '^output_detections_file$'
    m = re.match(int_exp, type_str)
    if m:
        groups = m.groups()
        min = int(groups[0]) if groups[0] is not None else None
        max = int(groups[1]) if groups[1] is not None else None
        return ConfigIntType(min, max)
    m = re.match(float_exp, type_str)
    if m:
        groups = m.groups()
        min = float(groups[0]) if groups[0] is not None else None
        max = float(groups[1]) if groups[1] is not None else None
        return ConfigFloatType(min, max)

    m = re.match(out_imagelist_exp, type_str)
    if m:
        return ConfigOutputImageListType()
    m = re.match(out_detections_exp, type_str)
    if m:
        return ConfigOutputDetectionsType()

    return StringType()


#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from typing import Dict, Tuple, List, Optional

from pep_tk.core.parser import VIAMEDataset
from pep_tk.core.configuration.types import parse_type
from pep_tk.core.configuration.exceptions import *
from pep_tk.core.configuration import ENV_VARIABLE, VALUE


class ConfigOption:
    def __init__(self, name: str, option_dict: Dict):
        self.name = name
        self.default = option_dict['default']
        self.type = option_dict['type']
        self.env_variable = option_dict['env_variable']
        self.description = option_dict.get('description')
        self.validator = parse_type(self.type)
        self.__value: Optional[VALUE] = option_dict.get('_value')
        self._locked = option_dict.get('_locked', False)

        # ensure default is valid
        valid, _ = self.validator.validate(self.default)
        if not valid:
            raise InvalidConfigOptionDefaultException('', self.name, self.default)

    def value(self) -> VALUE:
        if self.__value:
            return self.__value
        return self.default

    def set_value(self, value):
        """
        Sets the value of the ConfigOption
        :param value: new value for the config option
        :return: boolean
        """
        # value = str(value)
        if self._locked:
            return False

        valid, res = self.validator.validate(value)
        if valid:
            self.__value = res
            return True
        return False

    def reset(self):
        """ Resets the ConfigOption to the default value """
        if not self._locked:
            self.__value = self.default

    def get_env(self) -> Tuple[ENV_VARIABLE, VALUE]:
        """ returns a tuple with the first item being the environment variable name and the second begin the value """
        return self.env_variable, self.value()

    def to_dict(self) -> Dict:
        return {'name': self.name,
                '_value': self.value(),
                '_locked': self._locked,
                'default': self.default,
                'type': self.type,
                'env_variable': self.env_variable,
                'description': self.description}



class ConfigOptionGroup:
    def __init__(self, group_name: str, config_dict: Dict, required_group: bool = True):
        self.required_group = required_group
        self.group_name = group_name
        self.options: List[ConfigOption] = []

        group_dict = config_dict.get(group_name)
        if group_dict is None and self.required_group:
            raise MissingConfigGroupException(self.group_name)
        elif group_dict is not None:
            # make a list of ConfigOptions that are in this group
            for name, attributes in group_dict.items():
                opt_type = attributes.get('type')
                if not self.validate_option_type(attributes.get('type')):
                    raise InvalidConfigOptionTypeException(self.group_name, name, opt_type)

                o = ConfigOption(name, attributes)
                self.options.append(o)

    def validate_option_type(self, opt) -> bool:
        msg = self.__class__.__name__ + ' must implement validate_option_type.'
        raise NotImplementedError(msg)
        return False

    def get_config_option(self, name) -> Optional[ConfigOption]:
        """
        :param name: option name to get
        :return: returns the ConfigOption with that name or None if an option with that name doesn't exist
        """
        for opt in self.options:
            if opt.name == name:
                return opt
        return None

    def set_config_option(self, name, value: VALUE) -> bool:
        opt = self.get_config_option(name)
        if opt is not None:
            return opt.set_value(value)
        return False

    def reset_config_option(self, name):
        opt = self.get_config_option(name)
        if opt is not None:
            opt.reset()

    def reset_all(self):
        for opt in self.options:
            opt.reset()

    def get_env_ports(self) -> Dict[ENV_VARIABLE, VALUE]:
        env_list = [opt.get_env() for opt in self.options]
        return {k: v for (k, v) in env_list}

    # serialization (load/save) capabilities
    def to_dict(self):
        return {opt.name: opt.to_dict() for opt in self.options}

    # object methods
    def __len__(self):
        return len(self.options)

    def __getitem__(self, name) -> ConfigOption:
        return self.get_config_option(name)


class PipelineOutputOptionGroup(ConfigOptionGroup):
    __image_list_type = 'output_image_list'
    __detections_csv_type = 'output_detections_csv'

    def __init__(self, config_dict: Dict):
        super().__init__('output_config', config_dict, required_group=False)

    def validate_option_type(self, opt) -> bool:
        return opt in [self.__image_list_type, self.__detections_csv_type]

    def get_image_list_options(self) -> List[ConfigOption]:
        opts = []
        for opt in self.options:
            if opt.type == self.__image_list_type:
                opts.append(opt)
        return opts

    def get_det_csv_options(self) -> List[ConfigOption]:
        opts = []
        for opt in self.options:
            if opt.type == self.__detections_csv_type:
                opts.append(opt)
        return opts

    def get_image_list_env_ports(self):
        env_list = [opt.get_env() for opt in self.get_image_list_options()]
        return {k: v for (k, v) in env_list}

    def get_det_csv_env_ports(self):
        env_list = [opt.get_env() for opt in self.get_det_csv_options()]
        return {k: v for (k, v) in env_list}

class PipelineParametersOptionGroup(ConfigOptionGroup):
    def __init__(self, config_dict: Dict):
        super().__init__('parameters_config', config_dict, required_group=False)

    def validate_option_type(self, opt) -> bool: return True


class DatasetPipelineEnvAdaptersGroup:
    def __init__(self, config_dict: Dict):
        self.__config = config_dict.get('dataset_pipeline_adapters')

    def get_env_ports(self, dataset: VIAMEDataset, missing_ok=False) -> Dict[ENV_VARIABLE, VALUE]:
        res = {}
        missing_ports = []
        for name, attributes in self.__config.items():
            attr = attributes['dataset_attribute']
            if dataset.get(attr) is None:
                missing_ports.append(attr)

            if attr in dataset:
                res[attributes['env_variable']] = dataset[attr]

        if not missing_ok and len(missing_ports) > 0:
            raise MissingPortsException(missing_ports, dataset.name)

        return res

    def to_dict(self):
        return self.__config


class PipelineConfig:
    def __init__(self, pipeline_name, config_dict: Dict):
        """
        A PipelineConfig is made up of two config groups, the first is parameters such as thresholds, and
        the second is the pipeline output such as output detections file and image lists.  It also has
        dataset ports, which given a dataset will connect any dataset specific parameters required by the pipeline.
        :param pipeline_name: name of the pipeline
        :param config_dict: dictionary of the given pipeline from the parsed pipeline manifest yaml file
        """
        from pep_tk import PLUGIN_PATH
        self.name = pipeline_name
        self.path = os.path.join(PLUGIN_PATH, os.path.normpath(config_dict['path']))
        self.directory = os.path.dirname(self.path)
        if not os.path.exists(self.path):
            raise Exception('pipeline %s does not exist' % self.path)
        self.parameters_group = PipelineParametersOptionGroup(config_dict)
        self.output_group = PipelineOutputOptionGroup(config_dict)
        self.dataset_ports = DatasetPipelineEnvAdaptersGroup(config_dict)

    def get_output_env_ports(self, output_directory = ''):
        output_ports = self.output_group.get_env_ports()
        # append output directory to all output files
        output_ports_in_out_dir = {k: os.path.join(output_directory, v) for k,v in output_ports.items()}
        return output_ports_in_out_dir

    def get_parameter_env_ports(self):
        return self.parameters_group.get_env_ports()

    def get_pipeline_dataset_environment(self, dataset: VIAMEDataset, missing_ok=False) -> Dict[ENV_VARIABLE, VALUE]:
        return self.dataset_ports.get_env_ports(dataset, missing_ok)


    def to_dict(self) -> Dict:
        d = {
                'name': self.name,
                'parameters_config': self.parameters_group.to_dict(),
                'output_config': self.output_group.to_dict(),
                'dataset_pipeline_adapters': self.dataset_ports.to_dict(),
                'path': self.path
            }
        return d



#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

# https://github.com/PySimpleGUI/PySimpleGUI/issues/3058
import abc
import os
import PySimpleGUI as sg
from dataclasses import dataclass

from pep_tk.core.parser.load_dataset import safe_load_dataset_manifest
from pep_tk.psg.settings import get_user_settings, SystemSettingsNames, get_viame_bash_or_bat_file_path, UserProperties


def xstr(s):
    if s is None:
        return ''
    return str(s)


class Validator(metaclass=abc.ABCMeta):
    input_key = None

    @abc.abstractmethod
    def validate_ui(self, window: sg.Window) -> bool:
        pass

    @abc.abstractmethod
    def validate(self, v) -> bool:
        pass

    def value(self, window: sg.Window):
        return os.path.normpath(xstr(window[self.input_key].get()))

    def update_error(self, window: sg.Window, message=None):
        if message:
            window[self.input_key].update(text_color='red')
            sg.popup_no_buttons(message, title="Properties Error", keep_on_top=True, modal=True,
                                location=window.current_location())
        else:
            window[self.input_key].update(text_color='black')
        return False


class DataManifestValidator(Validator):
    input_key = '-dataset_manifest_filepath-IN-'

    def validate_ui(self, window: sg.Window) -> bool:
        if window[self.input_key].get() == "":
            return False
        dm, error = safe_load_dataset_manifest(self.value(window))
        self.update_error(window, error)
        return error is None

    def validate(self, v) -> bool:
        if v is None or v == '':
            return False
        dm, error = safe_load_dataset_manifest(v)
        return error is None


class JobBaseDirValidator(Validator):
    input_key = '-job_dir-IN-'

    def validate_ui(self, window: sg.Window) -> bool:
        if window[self.input_key].get() == "":
            return False
        error = None
        val = self.value(window)
        if not os.path.isdir(val):
            error = f'Job directory "{val}" does not exist.'
        self.update_error(window, error)
        return error is None

    def validate(self, v) -> bool:
        if v is None or v == '':
            return False
        return os.path.isdir(v)


class VIAMEDirValidator(Validator):
    input_key = '-setup_viame_filepath-IN-'

    def validate_ui(self, window: sg.Window) -> bool:
        if window[self.input_key].get() == "":
            return False
        error = None
        setup_viame_fp = get_viame_bash_or_bat_file_path(self.value(window))
        if not os.path.isfile(setup_viame_fp):
            error = f'Invalid VIAME/SEAL-TK directory.  ' \
                    f'Was unable to find setup_viame.sh(linux) or setup_viame.bat(windows) in the selected directory ' \
                    f'{setup_viame_fp}.',

        self.update_error(window, error)
        return error is None

    def validate(self, v) -> bool:
        if v is None or v == '':
            return False
        setup_viame_fp = get_viame_bash_or_bat_file_path(v)
        return os.path.isfile(setup_viame_fp)


dm_validator = DataManifestValidator()
viame_validator = VIAMEDirValidator()
jobdir_validator = JobBaseDirValidator()

@dataclass
class PropertiesWindowOutput:
    dm_valid: bool
    job_valid: bool
    viame_valid: bool
    properties_updated: bool = None

    @property
    def all_valid(self):
        return all([self.dm_valid, self.job_valid, self.viame_valid])

def check_inputs(window: sg.Window, update=True) -> PropertiesWindowOutput:
    d_valid = dm_validator.validate_ui(window)
    v_valid = viame_validator.validate_ui(window)
    j_valid = jobdir_validator.validate_ui(window)
    if update:
        user_settings = get_user_settings()
        if d_valid:
            user_settings[SystemSettingsNames.dataset_manifest_filepath] = dm_validator.value(window)

        if j_valid:
            user_settings[SystemSettingsNames.job_directory] = jobdir_validator.value(window)

        if v_valid:
            user_settings[SystemSettingsNames.viame_directory] = viame_validator.value(window)

    return PropertiesWindowOutput(dm_valid=d_valid, viame_valid=v_valid, job_valid=j_valid)


def check_settings():
    p = UserProperties()
    d_valid = dm_validator.validate(p.data_manifest_filepath)
    v_valid = viame_validator.validate(p.viame_dir)
    j_valid = jobdir_validator.validate(p.job_base_dir)
    out = PropertiesWindowOutput(dm_valid=d_valid, job_valid=j_valid, viame_valid=v_valid)
    return out

def show_properties_window(skip_if_valid=False, modal=True) -> PropertiesWindowOutput:
    """

    :param skip_if_valid:
    :param modal:
    :param error_message:
    :return: true if settings were updated, false if not
    """
    out = check_settings()
    if out.all_valid and skip_if_valid:
        out.properties_updated = False
        return out

    p = UserProperties()
    settings_before = p.as_dict()
    def properties_changed() -> bool:
        p.refresh()
        return p.as_dict() != settings_before

    manifest_folder = os.path.dirname(p.data_manifest_filepath) if p.data_manifest_filepath else None
    layout = [[sg.Text('Enter the viame directory:')],
              [sg.Input(xstr(p.viame_dir), key=VIAMEDirValidator.input_key),
               sg.FolderBrowse(initial_folder=p.viame_dir)],

              [sg.Text('Enter the dataset manifest filepath:')],
              [sg.Input(xstr(p.data_manifest_filepath), key=DataManifestValidator.input_key),
               sg.FileBrowse(initial_folder=manifest_folder)],

              [sg.Text('Enter the job base directory(directory where all jobs go):')],
              [sg.Input(xstr(p.job_base_dir), key=JobBaseDirValidator.input_key),
               sg.FolderBrowse(initial_folder=p.job_base_dir)],

              [sg.B('Complete Setup'), sg.B('Exit', key='Exit')]]

    user_settings = get_user_settings()
    location = user_settings[SystemSettingsNames.window_location] or (0, 0)
    window = sg.Window('PEP-TK: Properties',
                       layout,
                       keep_on_top=True,
                       finalize=True,
                       location=location,
                       modal=modal)

    while True:
        event, values = window.read()
        if event in (sg.WINDOW_CLOSED, 'Exit'):
            out = check_settings()
            out.properties_updated = False
            window.close()
            return out
        elif event == 'Complete Setup':
            out = check_inputs(window)
            out.properties_updated = properties_changed()
            if out.all_valid:
                window.close()
                return out




#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from configparser import ConfigParser, DuplicateSectionError, Error
from typing import List, Optional

import re

from pep_tk.core.parser import DatasetManifestError, DuplicateDatasetName, ImageListMissingImage, DatasetFileNotFound, ManifestParser, path_to_absolute
from pep_tk.core.parser import VIAMEDataset


class INIDatasetsParser(ConfigParser, ManifestParser):
    def read(self, filenames, encoding=None):
        if isinstance(filenames, (str, os.PathLike)):
            filenames = [filenames]
        read_ok = []
        for filename in filenames:
            try:
                with open(filename, encoding=encoding) as fp:
                        self._read(fp, filename)
            except OSError:
                continue
            except DuplicateSectionError as e:
                raise DuplicateDatasetName(e.message)
            except Error as e:
                raise DatasetManifestError(e.message)

            if isinstance(filename, os.PathLike):
                filename = os.fspath(filename)
            read_ok.append(filename)
            self.validate_dataset_files(filename)

        return read_ok

    def as_dict(self):
        d = dict(self._sections)
        for k in d:
            d[k] = dict(self._defaults, **d[k])
            d[k].pop('__name__', None)
        return d


    def validate_dataset_files(self, manifeset_fp):
        manifest_wd = os.path.dirname(manifeset_fp)
        manifeset_fn = os.path.basename(manifeset_fp)

        datasets = self.as_dict()
        for ds_name, attrs in datasets.items():
            # dataset must have an image list
            if self.att_thermal_image_list not in attrs and self.att_color_image_list not in attrs:
                raise Exception(f'[{manifeset_fn}][{ds_name}] ERROR: No color or a thermal image list defined.')

            for a, v in attrs.items():
                datafile_abspath = path_to_absolute(manifest_wd, v)
                datafile_wd = os.path.dirname(datafile_abspath)

                # check that dataset file was found
                if not os.path.isfile(datafile_abspath):
                    raise DatasetFileNotFound(f'[{manifeset_fn}][{ds_name}] ERROR: File "{a}={v}" does not exist.')

                # check that all images exist in the defined image list
                if a in [self.att_color_image_list, self.att_thermal_image_list]:
                    with open(datafile_abspath, 'r') as f:
                        image_paths = list(line for line in (l.strip() for l in f.readlines()) if line)
                    for img_fp in image_paths:
                        img_fp = path_to_absolute(datafile_wd, img_fp)
                        if not os.path.isfile(img_fp):
                            raise ImageListMissingImage(f'[{manifeset_fn}][{ds_name}] ERROR: "{img_fp}" was not found in {a}.')

                # set path to the absolute path
                self.set(ds_name,a, datafile_abspath)

    # given a string, list the dataset keys containing that string
    def list_dataset_keys_txt(self, txt: str) -> List[str]:
        try:
            regkey = '^.*' + txt + '.*$'
            r = re.compile(regkey)
            return list(filter(r.match, self.list_dataset_keys()))
        except:
            return []

    def list_dataset_keys(self) -> List[str]:
        return list(self._sections.keys())

    def get_dataset(self, name: str) -> Optional[VIAMEDataset]:
        ds = self.as_dict().get(name)
        if not ds:
            return None
        return VIAMEDataset(name=name,
                            color_image_list=ds.get(self.att_color_image_list),
                            thermal_image_list=ds.get(self.att_thermal_image_list),
                            transformation_file=ds.get(self.att_transform))



#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from typing import List, Optional
import re
import pandas as pd

from pep_tk.core.parser import VIAMEDataset
from pep_tk.core.parser import DuplicateDatasetName, MisingDatasetNameException, ImageListMissingImage, \
    NoImageListException, DatasetFileNotFound, ManifestParser, path_to_absolute



class CSVDatasetsParser(ManifestParser):
    attr_dataset_name = 'dataset_name'

    def __init__(self):
        self._datasets = {}

    def read(self, filename, fullcheck=False):
        cols = [self.attr_dataset_name, self.att_thermal_image_list, self.att_color_image_list,
                                self.att_transform]
        df = pd.read_csv(filename, comment='#', header=0,
                         usecols=cols)[cols]
        df = df.where(pd.notnull(df), None)
        self.validate_dataset_files(filename, df, fullcheck)

    def validate_dataset_files(self, manifeset_fp: str, df: pd.DataFrame, fullcheck):
        read_datasets = {}
        for i, row in df.iterrows():
            ds_name = row.get(self.attr_dataset_name)
            if not ds_name:
                raise MisingDatasetNameException(f'row {i} in {manifeset_fp} does not have a {self.attr_dataset_name}.')
            if ds_name in read_datasets or ds_name in self._datasets:
                raise DuplicateDatasetName(f'Duplicate {self.attr_dataset_name} found "{ds_name}"')
            read_datasets[ds_name] = {self.att_color_image_list: row[self.att_color_image_list],
                                      self.att_thermal_image_list: row[self.att_thermal_image_list],
                                      self.att_transform: row[self.att_transform]}

        manifest_wd = os.path.dirname(manifeset_fp)
        manifeset_fn = os.path.basename(manifeset_fp)

        for ds_name, attrs in read_datasets.items():
            # dataset must have an image list
            if not attrs.get(self.att_thermal_image_list) and not attrs.get(self.att_color_image_list):
                raise NoImageListException(f'[{manifeset_fn}][{ds_name}] ERROR: No color or a thermal image list defined.')

            for a, v in attrs.items():
                if v == None: continue
                datafile_abspath = path_to_absolute(manifest_wd, v)
                datafile_wd = os.path.dirname(datafile_abspath)

                # check that dataset file was found
                if not os.path.isfile(datafile_abspath):
                    raise DatasetFileNotFound(f'[{manifeset_fn}][{ds_name}] ERROR: File "{a}={v}" does not exist.')

                if fullcheck:
                    # check that all images exist in the defined image list
                    if a in [self.att_color_image_list, self.att_thermal_image_list]:
                        with open(datafile_abspath, 'r') as f:
                            image_paths = list(line for line in (l.strip() for l in f.readlines()) if line)
                        for img_fp in image_paths:
                            img_fp = path_to_absolute(datafile_wd, img_fp)
                            if not os.path.isfile(img_fp):
                                raise ImageListMissingImage(
                                    f'[{manifeset_fn}][{ds_name}] ERROR: "{img_fp}" was not found in {a}.')

                # set path to the absolute path
                read_datasets[ds_name][a] = datafile_abspath
        self._datasets.update(read_datasets)

    # given a string, list the dataset keys containing that string
    def list_dataset_keys_txt(self, txt: str) -> List[str]:
        try:
            regkey = '^.*' + txt + '.*$'
            r = re.compile(regkey)
            return list(filter(r.match, self.list_dataset_keys()))
        except:
            return []

    def list_dataset_keys(self) -> List[str]:
        return list(self._datasets.keys())

    def get_dataset(self, name: str) -> Optional[VIAMEDataset]:
        ds = self._datasets.get(name)
        if not ds:
            return None
        return VIAMEDataset(name=name,
                            color_image_list=ds.get(self.att_color_image_list),
                            thermal_image_list=ds.get(self.att_thermal_image_list),
                            transformation_file=ds.get(self.att_transform))



#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import abc
import PySimpleGUI as sg

from pep_tk.psg.settings import image_resource_path


class LayoutSection(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def get_layout(self):
        return

    @abc.abstractmethod
    def handle(self, window, event, values):
        return

    @property
    @abc.abstractmethod
    def layout_name(self) -> str:
        return ''


# SYMBOL_UP = '▲'
# SYMBOL_DOWN = '▼'
#
#
# def collapse(layout, key):
#     """
#     Helper function that creates a Column that can be later made hidden, thus appearing "collapsed"
#     :param layout: The layout for the section
#     :param key: Key used to make this seciton visible / invisible
#     :return: A pinned column that can be placed directly into your layout
#     :rtype: sg.pin
#     """
#     return sg.pin(sg.Column(layout, key=key))

def Collapsible(layout, key, title='', arrows=(sg.SYMBOL_DOWN, sg.SYMBOL_UP), collapsed=False):
    """
    User Defined Element
    A "collapsable section" element. Like a container element that can be collapsed and brought back
    :param layout:Tuple[List[sg.Element]]: The layout for the section
    :param key:Any: Key used to make this section visible / invisible
    :param title:str: Title to show next to arrow
    :param arrows:Tuple[str, str]: The strings to use to show the section is (Open, Closed).
    :param collapsed:bool: If True, then the section begins in a collapsed state
    :return:sg.Column: Column including the arrows, title and the layout that is pinned
    """
    return sg.Column([[sg.T((arrows[1] if collapsed else arrows[0]), enable_events=True, k=key+'-BUTTON-'),
                       sg.T(title, enable_events=True, key=key+'-TITLE-')],
                      [sg.pin(sg.Column(layout, key=key, visible=not collapsed, metadata=arrows))]], pad=(0,0))


def Menubar(menu_def, text_color, background_color, pad=(0, 0)):
    """
    A User Defined element that simulates a Menu element by using ButtonMenu elements
    :param menu_def: A standard PySimpleGUI menu definition
    :type menu_def: List[List[Tuple[str, List[str]]]
    :param text_color: color for the menubar's text
    :type text_color:
    :param background_color: color for the menubar's background
    :type background_color:
    :param pad: Amount of padding around each menu entry
    :type pad:
    :return: A column element that has a row of ButtonMenu buttons
    :rtype: sg.Column
    """
    row = []
    for menu in menu_def:
        text = menu[0]
        if sg.MENU_SHORTCUT_CHARACTER in text:
            text = text.replace(sg.MENU_SHORTCUT_CHARACTER, '')
        if text.startswith(sg.MENU_DISABLED_CHARACTER):
            disabled = True
            text = text[len(sg.MENU_DISABLED_CHARACTER):]
        else:
            disabled = False
        row += [sg.ButtonMenu(text, menu, border_width=0, button_color=f'{text_color} on {background_color}',key=text, pad=pad, disabled=disabled)]

    return sg.Column([row], background_color=background_color, pad=(0,0), expand_x=True)

def help_icon(tooltip, key=None):
    if key is None:
        key = str(hash(tooltip))
    return sg.Image(image_resource_path('help.png'), tooltip=tooltip, key=key, size=(16, 16))


#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
from typing import List, Optional, Dict

import PySimpleGUI as sg

from pep_tk.core.job import TaskStatus
from pep_tk.psg.events import ProgressGUIEventData
from pep_tk.psg.fonts import Fonts
from pep_tk.psg.layouts import LayoutSection
from pep_tk.psg.settings import icon_filepath
from pep_tk.psg.utils import set_pep_theme

set_pep_theme(sg)

def task_status_update_key(task_key):
    return f'--task-update-{task_key}--'


# Status icons to display in the tabs for each task
status_icons = {TaskStatus.SUCCESS: icon_filepath('success.png'),
                TaskStatus.ERROR: icon_filepath('error.png'),
                TaskStatus.RUNNING: icon_filepath('running.png'),
                TaskStatus.INITIALIZED: icon_filepath('pending.png'),
                TaskStatus.CANCELLED: icon_filepath('cancelled.png')}


class TaskTab(LayoutSection):
    """
        A better progress bar than the default PySimpleGUI Progress Bar
        - Worker thread triggers a window even, passing a ProgressGUIEventData object to the GUI with
          the task_progress_update_key as the event key.
        - This then updates the BetterProgressBar with that task_progress_update_key
        - Contains information on elapsed time, a progress counter, avg time/iteration, and task name
     """
    def __init__(self, task_key):
        self.task_key = task_key

        # globally used key for triggering an update for a task's BetterProgressBar GUI layout
        self.task_progress_update_key = task_status_update_key(self.task_key)

        # GUI element keys
        self._text_title_key = f'--tt-text-title-{self.task_key}--'
        self._pb_key = f'--tt-progress-bar-{self.task_key}--'
        self._elapsed_key = f'--tt-elapsed-{self.task_key}--'
        self._counter_key = f'--tt-counter-{self.task_key}--'
        self._iteration_time_key = f'--tt-iteration-time-{self.task_key}--'
        self._status_key = f'--tt-status-{self.task_key}--'
        self._output_files_key = f'--tt-output-files-{self.task_key}--'
        self._kwiver_output_key = f'--tt-kwiver-output-{self.task_key}--' + sg.WRITE_ONLY_KEY

        # GUI Button events
        self._cancel_event_key = f'--cancel-{self.task_key}--'


        self.images = {}
        self.max_count = 0
        self.progress_count = 0
        self.is_cancelled = False
        self.status = None
        self.completed_on_load = None

    def get_layout(self):
        def empty_string(s):
            return ' ' * len(s)

        status_icon = sg.Image(size=(24, 24), key=self._status_key)
        title = sg.T(self.task_key, size=(len(self.task_key), 1), key=self._text_title_key, font=Fonts.title_medium)
        pb = sg.ProgressBar(100, orientation='hs', size=(20, 4), key=self._pb_key)
        elapsed_str = empty_string('00:00:00 elapsed 00:00:00 remaining')
        time_elapsed = sg.T(elapsed_str, key=self._elapsed_key)
        counter_str = empty_string('00000/00000')
        counter = sg.T(counter_str, key=self._counter_key)
        iter_str = empty_string('x.xx seconds/iter')
        avg_iteration_time = sg.T(iter_str, key=self._iteration_time_key, size=(len('x.xx seconds/iter'), 1))
        output_files = sg.Column([[]], key=self._output_files_key)
        cancel_button = sg.Button('Cancel', key=self._cancel_event_key, disabled=True)
        output_title = sg.T("Output Log", font=Fonts.title_small)
        kwiver_output = sg.Multiline( key=self._kwiver_output_key,
                                      autoscroll=False, auto_refresh=True, disabled=True, expand_x=True, expand_y=True, size=(50,10))
        layout = [[status_icon, title, pb, time_elapsed, counter, avg_iteration_time],
                  [output_files],
                  [cancel_button],
                  [output_title],
                  [kwiver_output]]
        return layout

    def handle(self, window, event, values):
        # cancel button clicked
        if event == self._cancel_event_key or self.is_cancelled:
            self.is_cancelled = True
            return

        # if a progress event for any of the tasks comes through
        if event != self.task_progress_update_key:
            return

        progress: ProgressGUIEventData = values[event]
        if not self.completed_on_load:
            self.completed_on_load = progress.completed_on_load
        self._update_avg_iteration_time(window, progress.time_per_count)
        self._update_time_elapsed(window, progress.elapsed_time, progress.estimated_time_remaining)
        self._update_pb(window, progress.progress_count, progress.max_count)
        self._update_counter(window, progress.progress_count, progress.max_count)
        self._update_status(window, progress.task_status)
        self._update_output_files(window, progress.output_files)
        self._update_kwiver_output(window, progress.output_log)

        self.max_count = progress.max_count
        self.progress_count = progress.progress_count

    def _update_status(self, window: sg.Window, status: TaskStatus):
        if status is None:
            return
        self.status = status
        icon_fp = status_icons.get(status, None)
        window[self._status_key].update(filename=icon_fp)

        if status == TaskStatus.RUNNING:
            window[self._cancel_event_key](disabled=False)
        else:
            window[self._cancel_event_key](disabled=True)

    def _update_kwiver_output(self, window: sg.Window, new_str: str):
        if new_str is None or new_str == "":
            return
        window[self._kwiver_output_key].print(new_str.strip(), autoscroll=True)

    def _update_avg_iteration_time(self, window: sg.Window, avg_iteration_time: float):
        if avg_iteration_time is None:
            return
        fmt = '%.2f seconds/iter' % avg_iteration_time
        window[self._iteration_time_key](value=fmt)

    def _update_time_elapsed(self, window: sg.Window, elapsed_time: float, remaining_time: float):
        if elapsed_time is None or remaining_time is None:
            return
        elapsed = str(datetime.timedelta(seconds=int(elapsed_time)))
        remaining = str(datetime.timedelta(seconds=int(remaining_time)))
        v = '%s elapsed %s remaining' % (elapsed, remaining)
        window[self._elapsed_key](value=v)

    def _update_pb(self, window: sg.Window, count: int, max_count: int):
        if count is None or max_count is None: return
        window[self._pb_key].update_bar(count, max_count)

    def _update_counter(self, window: sg.Window, count: int, max_count: int):
        if count is None or max_count is None: return
        window[self._counter_key](value='%d/%d' % (count, max_count))

    def _update_output_files(self, window: sg.Window, output_files: Optional[List[str]]):
        if not output_files or len(output_files) == 0:
            return
        new_elems = [[sg.T('Output Files:', font=Fonts.title_small)]]
        w = max([len(x) for x in output_files])
        new_elems.append([sg.Multiline('\n'.join(output_files), disabled=True, size=(w,len(output_files)))])
        window.extend_layout(window[self._output_files_key], new_elems)

    def layout_name(self) -> str:
        return self.task_progress_update_key


class TaskRunnerTab():
    def __init__(self, layout, task_key, visible):
        self._layout = layout
        self.task_key = task_key
        self.tab_button_key = f'-tab-button-{self.task_key}-'
        self.tab_contents_key = f'-tab-contents-{self.task_key}-'
        self.tab_status_image_key = f'-tab-status-image-{self.task_key}-'
        self.tab_status_update_key = task_status_update_key(self.task_key)
        self.tab_key = f'--tab--{self.task_key}--'
        self.visible = visible

    def get_layout(self):
        return sg.Column(self._layout, visible=self.visible, key=self.tab_contents_key, expand_y=True)

    def update_status(self, window: sg.Window, status: TaskStatus):
        icon_fp = status_icons.get(status, None)
        if icon_fp is None: return
        window[self.tab_status_image_key].Update(filename=icon_fp)


class TaskRunnerTabGroup(LayoutSection):
    button_color_off = ('black', '#eff1f3') # (Text Color, Button Color)
    button_color_on = ('black', 'azure')

    def __init__(self, items):
        tabs = []
        self.current_tab = None
        for tab in items:
            t = TaskRunnerTab(tab[0], tab[1], visible=self.current_tab is None)
            tabs.append(t)
            if self.current_tab is None:
                self.current_tab = t
        self.tabs_by_task_key : Dict[str, TaskRunnerTab] = {t.task_key: t for t in tabs}
        self.tab_button_event_keys : Dict[str, TaskRunnerTab] = {t.tab_button_key: t for t in tabs}
        self.tab_contents_event_keys : Dict[str, TaskRunnerTab] = {t.tab_contents_key: t for t in tabs}
        self.update_event_keys : Dict[str, TaskRunnerTab] = {t.tab_status_update_key: t for t in tabs}
        self._tasks_started_flags = []

    def get_layout(self):
        # Calculate sizings
        icon_dim = 24
        max_name = 0
        max_allowable_name_len = 100
        for t in self.tabs_by_task_key.values():
            if len(t.task_key) > max_name:
                max_name = len(t.task_key)
        max_name = min(max_name, max_allowable_name_len)

        row_height = 30
        calculated_height = len(self.tabs_by_task_key.values()) * row_height
        min_height, max_height = 20 * row_height, 30 * row_height
        height = min(max(min_height, calculated_height), max_height)

        calculated_width = int(max_name * 8.5)
        min_width, max_width = 10 * 9, max_allowable_name_len * 9
        btn_width = min(max(min_width, calculated_width), max_width)
        col_width = btn_width + icon_dim
        # Create layout elements
        tabs = []
        contents = []
        selected_name = ""
        def make_max_name_len(name):
            cur_len = len(name)
            spaces = " "* ((max_name - cur_len))
            return name + spaces

        for t in self.tabs_by_task_key.values():
            tab_col = \
                [sg.Frame('', layout=[[sg.Image(size=(icon_dim, icon_dim),
                                                key=t.tab_status_image_key,
                                                pad=((5, 5), (5, 5)),
                                                background_color=self.button_color_off[1]),
                                   sg.Button(make_max_name_len(t.task_key),
                                             key=t.tab_button_key,
                                             pad=((0, 0), (0, 0)),
                                             button_color=self.button_color_off,
                                             use_ttk_buttons=True,
                                             mouseover_colors=self.button_color_on,
                                             border_width=0)]], key = t.tab_key, background_color=self.button_color_off[1])]
            tabs.append(tab_col)
            contents.append(t.get_layout())

        needs_scroll = len(tabs) * row_height > height
        scrollable_tabs =sg.Column(tabs, scrollable=needs_scroll, vertical_scroll_only=True, size=(col_width, height),
                          background_color=self.button_color_off[1], pad=((5, 5), (5, 5)), vertical_alignment='top')


        tab_contents = sg.Frame(f'Task Progress: {selected_name}', [contents], vertical_alignment='top',
                 key='-progress-frame-')
        layout = [[scrollable_tabs, tab_contents]]
        return layout

    def select_tab(self, window, event_tab_btn_key = None):
        if event_tab_btn_key == self.current_tab.tab_button_key:
            return
        if event_tab_btn_key is None:
            event_tab_btn_key = self.current_tab.tab_button_key
        # Hide other current tab
        window[self.current_tab.tab_contents_key].Update(visible=False)
        window[self.current_tab.tab_key].Widget.config(background=self.button_color_off[1])
        window[self.current_tab.tab_button_key].Update(button_color=self.button_color_off)

        window[self.current_tab.tab_button_key].ParentRowFrame.config(background=self.button_color_off[1])
        window[self.current_tab.tab_button_key].ParentRowFrame.config(highlightbackground=self.button_color_off[1])
        window[self.current_tab.tab_button_key].ParentRowFrame.config(highlightcolor=self.button_color_off[1])#self.button_color_on[1])
        window[self.current_tab.tab_button_key].set_focus(False)

        # Make new tab visible
        tab = self.tab_button_event_keys[event_tab_btn_key]

        window[tab.tab_contents_key].Update(visible=True)
        window[tab.tab_button_key].Update(button_color=self.button_color_on)
        window[tab.tab_button_key].set_focus(True)
        window[tab.tab_contents_key].expand(True,True,True)

        window[tab.tab_button_key].ParentRowFrame.config(background=self.button_color_on[1])
        window[tab.tab_button_key].ParentRowFrame.config(highlightbackground=self.button_color_on[1])
        window[tab.tab_button_key].ParentRowFrame.config(highlightcolor=self.button_color_on[1])#self.button_color_on[1])

        window['-progress-frame-'].Update(value=f'Task Progress: {tab.task_key}')
        self.current_tab = tab

    def handle(self, window, event, values):
        if event in self.tab_button_event_keys:
            self.select_tab(window, event)

        elif event in self.update_event_keys:
            progress: ProgressGUIEventData = values[event]
            self.update_event_keys[event].update_status(window, progress.task_status)
            task_key = self.update_event_keys[event].task_key
            if progress.task_status == TaskStatus.RUNNING and task_key not in self._tasks_started_flags:
                self._tasks_started_flags.append(task_key)
                self.select_tab(window, self.update_event_keys[event].tab_button_key)

    @property
    def layout_name(self):
        return 'Tabs'



#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Tuple, Optional
import PySimpleGUI as sg

from pep_tk.psg.fonts import Fonts
from pep_tk.psg.settings import image_resource_path


def popup_error(msg: str, parent_window_loc: Optional[Tuple[int,int]] = None, parent_window_size: Optional[Tuple[int,int]] = None):
    if parent_window_loc is None or parent_window_size is None:
        sg.popup_ok(msg, title='Uh oh', keep_on_top=True)
    else:
        max_line_width = 200
        current = sg.MESSAGE_BOX_LINE_WIDTH
        for line in msg.split('\n'):
            if len(line) > current:
                current = len(line)

        (win_x, win_y) = parent_window_loc
        (win_w, win_h) = parent_window_size
        dim_multiplier = 7
        popup_w = dim_multiplier * min(max_line_width, current)
        popup_h = 50 + dim_multiplier * len(msg.split('\n'))
        cx = int(win_w / 2 - popup_w / 2)
        cy = int(win_h / 2 - popup_h / 2)
        sg.popup_ok(msg, title='Uh oh', line_width=popup_w, location=(win_x + cx, win_y + cy), keep_on_top=True)

def popup_about(location=(None,None)):
    from pep_tk import __version__
    message = f'PEP GUI Version {__version__}\n' \
              f'Developed by Yuval Boss\n' \
              f'https://github.com/readicculus/pep_gui'

    # sg.popup_no_buttons(message, title='About', line_width=200, location=location, modal=True, keep_on_top=True,
    #                      font=Fonts.tab_text)
    w=max([len(x) for x in message.split('\n')])
    l = [[sg.Column([[sg.Image(filename=image_resource_path('icon_160x160.png'))]], vertical_alignment='center', justification='center',  k='-C-')],
         [sg.Multiline(default_text=message, disabled=True, no_scrollbar=True, expand_y=True, size=(w,3))]]
    window = sg.Window(layout=l, title='About', modal=True, keep_on_top=True, location=location, font=Fonts.tab_text, finalize=True)
    while True:
        event, values = window.read()
        if event in (sg.WINDOW_CLOSED, 'Exit'):
            window.close()
            break



#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from typing import Optional, Tuple

from pep_tk.core.parser import CSVDatasetsParser, INIDatasetsParser, ParserNotFoundException
from pep_tk.core.parser import DatasetManifestError, ManifestParser



def load_dataset_manifest(manifest_fp) -> ManifestParser:
    """
    :returns a ManifestParser
    """
    if manifest_fp is None:
        raise DatasetManifestError('No Manifest given')
    if not os.path.isfile(manifest_fp):
        str = f'Dataset manifest file "{manifest_fp}" does not exist.'
        raise DatasetManifestError(str)

    if manifest_fp.endswith('.csv'):
        dm = CSVDatasetsParser()
    elif manifest_fp.endswith('.cfg') or manifest_fp.endswith('.ini'):
        dm = INIDatasetsParser()
    else:
        raise ParserNotFoundException(
            f'Invalid manifest file format.  Can take csv(.csv) format, or ini format (.ini or .cfg).\n'
            f'"{manifest_fp}"')
    dm.read(manifest_fp)
    return dm


def safe_load_dataset_manifest(manifest_fp) -> Tuple[Optional[ManifestParser], Optional[str]]:
    try:
        dm = load_dataset_manifest(manifest_fp)
        return dm, None
    except DatasetManifestError as e:
        if hasattr(e, 'message'):
            msg = f'Error Type: {e.__class__.__name__}\nMessage:\n{e.message}'
        else:
            msg = str(e)
        return None, msg
    except Exception as e:
        if hasattr(e, 'message'):
            msg = f'Error Type: {e.__class__.__name__}\nMessage:\n{e.message}'
        else:
            msg = str(e)
        return None, msg



#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, Optional

from pep_tk.core.configuration import PipelineManifest, ConfigOption, PipelineConfig
from pep_tk.psg.fonts import Fonts
from pep_tk.psg.layouts import LayoutSection
import PySimpleGUI as sg


def pipeline_key(pipeline_name):
    return f'-F-{pipeline_name}-'


class PipelineConfigLayout(LayoutSection):
    def __init__(self, pipeline_name, pipeline_manifest):
        self.pipeline_name = pipeline_name
        self.pipeline_manifest: PipelineManifest = pipeline_manifest
        self.selected_pipeline: PipelineConfig = self.pipeline_manifest[self.pipeline_name]
        self.event_keys = [self.reset_defaults_key, self.pipeline_frame_key]
        self.input_keys_to_config_name: Dict[str, str] = {}
        # self.current_errors = {}

    def get_layout(self):
        config_layout = []
        config_layout.append([sg.T(self.pipeline_name, font=Fonts.title_medium)])
        config_layout.append([sg.HorizontalSeparator(color='black')])
        opts = self.selected_pipeline.parameters_group.options
        for o in opts:
            k = self.pipeline_config_input_key(o.name)
            self.event_keys.append(k)
            self.input_keys_to_config_name[k] = o.name
            row = [
                sg.T(f'{o.name}:', font=Fonts.description_bold),
                sg.I(key=k, default_text=o.value(), enable_events=True, background_color='white'),
                sg.T(f'({o.validator.description()})', key=self.pipeline_config_warning_key(o.name), text_color='red',
                     visible=False)
            ]
            config_layout.append(row)
            config_layout.append(
                [
                    sg.T(o.description, font=Fonts.description, pad=((0, 0), (0, 0)))
                 ]
            )
            config_layout.append([sg.HorizontalSeparator(color='white')])
        config_layout.append([sg.Button('Reset Defaults', key=self.reset_defaults_key)])
        return config_layout

    def validate(self, values) -> bool:
        for key, opt_name in self.input_keys_to_config_name.items():
            opt = self.selected_pipeline.parameters_group.get_config_option(opt_name)
            ui_value = values[key]
            ok, res = opt.validator.validate(ui_value)
            if not ok:
                return False
        return True

    def handle(self, window, event, values):
        if event == self.reset_defaults_key:
            self.selected_pipeline.parameters_group.reset_all()
            for key, opt_name in self.input_keys_to_config_name.items():
                warning_key = self.pipeline_config_warning_key(opt_name)
                opt = self.selected_pipeline.parameters_group.get_config_option(opt_name)
                window[key].Update(value=opt.value())
                window[key].Update(background_color='White')
                window[warning_key].Update(visible=False)

        elif event in list(self.input_keys_to_config_name.keys()):
            for key, opt_name in self.input_keys_to_config_name.items():
                warning_key = self.pipeline_config_warning_key(opt_name)
                opt = self.selected_pipeline.parameters_group.get_config_option(opt_name)
                config_opt_value = opt.value()
                ui_value = values[key]
                if not str(config_opt_value) == ui_value or key == event:
                    ok, res = opt.validator.validate(ui_value)
                    if ok:
                        opt.set_value(ui_value)
                        window[key].Update(background_color='White')
                        window[warning_key].Update(visible=False)
                    else:
                        window[key].Update(background_color='Red')
                        window[warning_key].Update(visible=True)

    def layout_name(self) -> str:
        return self.pipeline_name

    def pipeline_frame_key(self):
        return pipeline_key(self.pipeline_name)

    def pipeline_config_input_key(self, config_name):
        return f'-IN-{self.pipeline_name}-{config_name}-'

    def pipeline_config_warning_key(self, config_name):
        return f'-WARN-{self.pipeline_name}-{config_name}-'

    @property
    def reset_defaults_key(self):
        return f'-IN-{self.pipeline_name}-reset-'

    def get_opt_from_key(self, k) -> ConfigOption:
        pass


class PipelineSelectionLayout(LayoutSection):
    def __init__(self, pipeline_manifest: PipelineManifest, event_key='_pipeline_tab_'):
        self.pipeline_manifest = pipeline_manifest
        self.selected_pipeline: Optional[PipelineConfig] = None

        # gui element identifier keys
        self.event_key = event_key
        self.combobox_key = '%s_combobox' % self.event_key
        self.reset_defaults_key = '%s_reset_defaults_button' % self.event_key

        # create pipeline config layouts
        self.config_frames = {}
        for p in self.pipeline_manifest.list_pipeline_names():
            layout = PipelineConfigLayout(p, self.pipeline_manifest)
            self.config_frames[layout.pipeline_frame_key()] = layout

    def validate(self, values) -> bool:
        if self.selected_pipeline is None:
            return False
        selected_pipeline_key = pipeline_key(self.selected_pipeline.name)
        return self.config_frames[selected_pipeline_key].validate(values)

    def get_layout(self):
        pipelines = self.pipeline_manifest.list_pipeline_names()
        width = len(max(pipelines, key=len)) + 5
        frames = []
        for frame_key, pipeline_layout in self.config_frames.items():
            l = pipeline_layout.get_layout()
            n = pipeline_layout.pipeline_name
            config_count = len(pipeline_layout.input_keys_to_config_name)
            if config_count <= 3:
                frames.append(sg.Column(
                    [[sg.Frame('', l, font=Fonts.title_small)]],
                    expand_x=True, key=frame_key,
                    visible=False))
            else:
                # put in a scrollable column if  more than 3 configs
                col = [[sg.Column(l, expand_x=True, scrollable=True, vertical_scroll_only=True)]]
                frames.append(sg.Frame('', col, key=frame_key, visible=False, font=Fonts.title_small))
        default_value = '<select a pipeline>'
        return [
            [sg.Text('Select and Configure a pipeline.', font=Fonts.description)],
            [sg.Combo(values=[default_value] + pipelines,
                      size=(width, min(10, len(pipelines) + 1)),
                      default_value=default_value,
                      key=self.combobox_key,
                      font=Fonts.description,readonly=True,
                      enable_events=True)
             ],
            frames
        ]

    def get_selected_pipeline(self):
        return self.selected_pipeline

    # event loop handler
    def handle(self, window, event, values):
        if event == self.combobox_key:
            selected_pipeline_name = values[self.combobox_key]
            if selected_pipeline_name == '<select a pipeline>' or selected_pipeline_name == '':
                if self.selected_pipeline is not None:
                    window[self.combobox_key](value= self.selected_pipeline.name)
                return
            if selected_pipeline_name == self.selected_pipeline:
                return

            self.selected_pipeline = self.pipeline_manifest[selected_pipeline_name]

            frame_key = pipeline_key(selected_pipeline_name)
            window[frame_key](visible=True)
            for p in self.pipeline_manifest.list_pipeline_names():
                if p != selected_pipeline_name:
                    window[pipeline_key(p)](visible=False)
        else:
            for pipe_layout in self.config_frames.values():
                if event in pipe_layout.event_keys:
                    pipe_layout.handle(window, event, values)
                    break

    @property
    def layout_name(self):
        return 'Pipeline Selection'



main_font = 'Times New Roman'
#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

# main_font = 'Helvetica'
main_font = 'Arial'
class Fonts:
    description = (main_font, 10)
    description_bold = (main_font, 11, 'bold')

    title_large = (main_font, 18)
    title_medium = (main_font, 16)
    title_small = (main_font, 12)
    tab_text = (main_font, 11)


#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os

import yaml

from pep_tk import PLUGIN_PATH
from pep_tk.core.configuration import PipelineConfig
default_manifest = os.path.normpath(os.path.join(PLUGIN_PATH, 'conf/pipeline_manifest.yaml'))

class PipelineManifest:
    __root = 'PipelineManifest'

    def __init__(self, manifest_file: str = default_manifest):
        self.manifest_file = manifest_file
        with open(manifest_file, 'r') as stream:
            try: dataset_yaml = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                print(exc)

        self.pipelines = {}
        for pipeline_name, pipeline_config in dataset_yaml[self.__root].items():
            self.pipelines[pipeline_name] = PipelineConfig(pipeline_name, pipeline_config)

    def list_pipeline_names(self):
        return list(self.pipelines.keys())

    def __getitem__(self, pipeline_name: str) -> PipelineConfig:
        return self.pipelines[pipeline_name]




#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import threading
import time
from typing import List

import PySimpleGUI as sg

from pep_tk.core.job import load_job, TaskStatus, TaskKey
from pep_tk.core.scheduler import Scheduler
from pep_tk.psg.events import GUIManager
from pep_tk.psg.fonts import Fonts
from pep_tk.psg.layouts import TaskTab, TaskRunnerTabGroup
from pep_tk.psg.settings import get_user_settings, SystemSettingsNames, get_viame_bash_or_bat_file_path
from pep_tk.psg.utils import set_pep_theme

set_pep_theme(sg)



def make_main_window(tasks: List[TaskKey], gui_settings: sg.UserSettings):
    tabs = [TaskTab(task_key) for task_key in tasks]
    tabs_group = TaskRunnerTabGroup([(t.get_layout(), t.task_key) for t in tabs])
    header = sg.Frame(title='', layout=[[sg.Text('Job Progress:', font=Fonts.title_medium)],
              [sg.Text('xxxxxxx/xxxxxxx images or XX.XX% complete.', key='--total-progress-ta--')]], expand_x=True)
    layout = [[header],
              [tabs_group.get_layout()]]

    location = gui_settings.get(SystemSettingsNames.window_location, (0, 0))
    window = sg.Window('PEP-TK: Job Runner', layout, location=location, finalize=True,
                       enable_close_attempted_event=True, use_default_focus=False)

    window['-progress-frame-'].expand(True, True, True)
    tabs_group.select_tab(window)  # ensure first tab is selected

    return window, tabs, tabs_group


def run_job(job_path: str):
    user_settings = get_user_settings()
    job_state, job_meta = load_job(job_path)

    window, tabs, tabs_group = make_main_window(job_state.tasks(), user_settings)
    tabs_by_update_key = {t.task_progress_update_key: t for t in tabs}
    manager = GUIManager(window=window, tabs=tabs)
    kill_event = threading.Event()
    sched = Scheduler(job_state=job_state,
                      job_meta=job_meta,
                      manager=manager,
                      kwiver_setup_path=get_viame_bash_or_bat_file_path(
                          user_settings.get(SystemSettingsNames.viame_directory)), kill_event=kill_event)

    threading.Thread(target=sched.run, daemon=True).start()

    def update_total_progress(window: sg.Window, start_time: int):
        total_progress = 0
        total = 0
        completed_on_load_count = 0
        for tab in tabs:
            total += tab.max_count
            if tab.status == TaskStatus.SUCCESS:
                total_progress += tab.max_count
                if tab.completed_on_load:
                    completed_on_load_count += tab.max_count
            else:
                total_progress += tab.progress_count

        total_completed_items = total_progress - completed_on_load_count
        elapsed_time = time.time() - start_time
        if total_completed_items == 0:
            time_per_epoch = 0
        else:
            time_per_epoch = elapsed_time / total_completed_items

        remaining_time = (total - total_progress) * time_per_epoch
        percent = (total_progress/total) * 100

        elapsed_str = str(datetime.timedelta(seconds=int(elapsed_time)))
        remaining_str = str(datetime.timedelta(seconds=int(remaining_time)))

        window['--total-progress-ta--'].update(value=f'{total_progress}/{total} items({percent:.2f}%) complete.  {time_per_epoch:.2f} seconds/iter.  '
                                                     f'Elapsed time {elapsed_str}. Estimated time remaining {remaining_str}.')

    start_time = time.time()
    while True:
        event, values = window.read()
        if event == sg.WIN_CLOSED:
            window.close()
            break
        elif event == sg.WIN_X_EVENT:
            # If user tries to close the window by clicking X, show a popup asking to confirm the action.
            try:
                location = window.current_location()
            except:
                location = (None, None)
            res = sg.popup_ok_cancel("Closing this window will stop any active tasks, "
                                     "are you sure you want to close?",
                                     title='Are you sure?', location=location)
            if res == 'OK':
                kill_event.set()
                window.close()
                break
        else:
            if event in tabs_by_update_key:
                tabs_by_update_key[event].handle(window, event, values)
                update_total_progress(window, start_time)

            tabs_group.handle(window, event, values)




#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

class MissingPortsException(Exception):
    def __init__(self, ports, dataset_name):
        self.ports = ports
        self.dataset_name = dataset_name
        self.message = 'This pipeline requires %s, which was not defined in the dataset %s.' % (str(self.ports), self.dataset_name)

        super().__init__(self.message)

class InvalidConfigOptionDefaultException(Exception):
    def __init__(self, group, name, default):
        self.group = group
        self.name = name
        self.default = default
        self.message = 'Config %s:%s has default defined as %s which is invalid.' % \
                       (self.group, self.name, str(self.default))
        super().__init__(self.message)

class InvalidConfigOptionTypeException(Exception):
    def __init__(self, category, name, type):
        self.category = category
        self.name = name
        self.type = type
        self.message = 'Config %s:%s type %s invalid.' % (self.category, self.name, self.type)
        super().__init__(self.message)

class MissingConfigGroupException(Exception):
    def __init__(self, group_name):
        self.group_name = group_name
        self.message = 'Config group is required and is not defined.' % (self.group_name)
        super().__init__(self.message)


#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
__version__ = "N/A"
PLUGIN_PATH = os.path.normpath(os.path.dirname(os.path.abspath(__file__)))

version_txt_fp = os.path.join(os.path.dirname(__file__), 'VERSION.txt')
with open(version_txt_fp, 'r') as file:
    v_data = file.read().replace('\n', '')
if len(v_data.split('.')) == 3:
    __version__ = v_data
else:
    print(f'Invalid VERSION.cfg - could not parse {v_data}')


#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from typing import Dict, Any

import PySimpleGUI as sg

from pep_tk.core.configuration import PipelineManifest
from pep_tk.core.configuration.exceptions import MissingPortsException
from pep_tk.core.job import create_job, job_exists
from pep_tk.core.parser import ManifestParser
from pep_tk.psg.fonts import Fonts
from pep_tk.psg.layouts import DatasetSelectionLayout, PipelineSelectionLayout, LayoutSection
from pep_tk.psg.settings import get_user_settings, SystemSettingsNames
from pep_tk.psg.utils import move_window_onto_screen
from pep_tk.psg.windows import show_properties_window, run_job, popup_error, popup_about


# ======== Handler helper functions =========
def validate_inputs(window: sg.Window, values: Dict[Any, Any], dataset_tab, pipeline_tab) -> bool:
    user_settings = get_user_settings()
    input_job_name = values['-job_name-IN-']
    combined_job_dir = os.path.join(user_settings.get(SystemSettingsNames.job_directory), input_job_name)
    input_datasets = dataset_tab.get_selected_datasets()
    input_pipeline = pipeline_tab.get_selected_pipeline()
    w_loc, w_size = window.current_location(), window.size
    # Check if no datasets were selected
    if len(input_datasets) < 1:
        popup_error('No datasets were selected.  Must select one or more datasets above.', w_loc, w_size)
        return False

    # if pipeline_tab.selected_pipeline is None:
    if not pipeline_tab.validate(values):
        popup_error('Either a pipeline isn\'t selected or error in configuration values.', w_loc, w_size)
        return False

    # Check for missing ports(aka if datasets/pipeline are not compatible)
    missing_ports = {}
    for dataset in input_datasets:
        try:
            input_pipeline.get_pipeline_dataset_environment(dataset)
        except MissingPortsException as e:
            missing_ports[e.dataset_name] = e.ports
    if len(missing_ports) > 0:
        msg = "Datasets aren't compatible with the selected pipeline: \n"
        for dataset_name, ports in missing_ports.items():
            msg += "%s: MISSING(%s)\n" % (dataset_name, ', '.join(ports))
        popup_error(msg, w_loc, w_size)
        return False

    # Check if the selected name is an empty string
    if input_job_name == '':
        popup_error('No job name entered', w_loc, w_size)
        return False

    # Check if the job directory(within the base directory) already exists
    if os.path.isdir(combined_job_dir):
        popup_error(f'Job {input_job_name} already exists, cannot override an existing job.\n{combined_job_dir}',
                    w_loc, w_size)
        return False

    return True


# ======== Create Job Window launcher =========
def launch_gui(pm: PipelineManifest, dm: ManifestParser) -> bool:
    sg.theme('SystemDefaultForReal')

    # ======== Page sections =========
    dataset_tab = DatasetSelectionLayout(dm)
    pipeline_tab = PipelineSelectionLayout(pm)

    def create_frame(tl: LayoutSection):
        return sg.Frame(layout=tl.get_layout(), title=tl.layout_name, font=Fonts.title_medium, title_color='#0b64c5')

    # ======== Create the Layout =========
    menu_def = [['&File', ['&Resume Job     Ctrl-R::-resume-menu-btn-', '&Properties::-properties-menu-btn-',
                           'E&xit::-exit-menu-btn-']],
                ['&Tools', '&Validate Datasets::-validate-menu-btn-'],
                ['&Help', '&About...::-about-menu-btn-'], ]

    layout = [
        [sg.Menu(menu_def, tearoff=False, pad=(200, 1))],
        [sg.Text('Polar Ecosystems Program Batch Runner', size=(38, 1), justification='center', font=Fonts.title_large,
                 relief=sg.RELIEF_RIDGE, k='-TEXT HEADING-', enable_events=True, text_color='#063970')],
        [create_frame(dataset_tab)],
        [create_frame(pipeline_tab)],
        [sg.Text('Job Name', font=Fonts.description), sg.Input('', key='-job_name-IN-', size=(20, 1))],
        [sg.Button('Create Job', key='-CREATE_JOB-')]]

    user_settings = get_user_settings()
    location = user_settings.get(SystemSettingsNames.window_location, (None, None))
    window = sg.Window('PEP-TK: Job Configuration', layout,
                       default_element_size=(12, 1), location=location, finalize=True)
    # move back on screen if off screen for some reason
    move_window_onto_screen(window)

    # ======== Show Properties window if properties are not valid =========
    out = show_properties_window(skip_if_valid=True)
    if out.all_valid and out.properties_updated:
        window.close()
        return False  # Reload GUI if properties changed
    if not out.all_valid and not out.properties_updated:
        window.close()
        return True  # Don't reload GUI if user doesn't fix properties and exits out

    # ======== Window / Event loop =========
    CREATED_JOB_PATH = None
    RESUME_JOB_PATH = None
    while True:
        event, values = window.read(timeout=2000)
        if event == sg.WIN_CLOSED:  # always,  always give a way out!
            break
        try:
            user_settings.set(SystemSettingsNames.window_location, window.CurrentLocation())
        except:
            pass
        if event == "__TIMEOUT__":
            continue
        # ======== Handle menu button pressed =========
        if '::' in event:
            menu_event = event.split('::')[1]  # event
            if menu_event == '-resume-menu-btn-':
                initial_folder = user_settings[SystemSettingsNames.job_directory]
                location = window.current_location()
                job_folder = sg.popup_get_folder('Select the job directory',
                                                 no_window=True,
                                                 keep_on_top=True,
                                                 initial_folder=initial_folder,
                                                 modal=True,
                                                 location=location)
                if job_folder:
                    exists = job_exists(job_folder)
                    if exists:
                        RESUME_JOB_PATH = job_folder
                        break
                    else:
                        popup_error(f'Job {job_folder} is not a valid job directory.', window.current_location(),
                                    window.size)
            elif menu_event == '-properties-menu-btn-':
                out = show_properties_window(modal=True)
                if out.properties_updated:
                    window.close()
                    return False  # Reload GUI
            elif menu_event == '-exit-menu-btn-':
                break  # exit loop
            elif menu_event == '-about-menu-btn-':
                popup_about(location=window.current_location())
            elif menu_event == '-validate-menu-btn-':
                pass

        # ======== Handle Create Job button pressed =========
        elif event == '-CREATE_JOB-':
            if validate_inputs(window, values, dataset_tab, pipeline_tab):
                selected_job_directory = user_settings.get(SystemSettingsNames.job_directory)
                selected_job_name = values['-job_name-IN-']

                pipeline = pipeline_tab.get_selected_pipeline()
                datasets = dataset_tab.get_selected_datasets()
                try:
                    job_dir = os.path.join(selected_job_directory, selected_job_name)
                    CREATED_JOB_PATH = create_job(pipeline=pipeline, datasets=datasets, directory=job_dir)
                except Exception as e:
                    popup_error(
                        f'There was an error creating the job: \n {str(e)}.\n I would recommend sending this error to Yuval.',
                        window.current_location(), window.size)
                    continue

                break  # END: close window
        else:
            # ======== Handle other user interactions =========
            dataset_tab.handle(window, event, values)
            pipeline_tab.handle(window, event, values)

    window.close()

    if CREATED_JOB_PATH:
        run_job(CREATED_JOB_PATH)
    elif RESUME_JOB_PATH:
        run_job(RESUME_JOB_PATH)

    return True



"""
Original: https://github.com/SzieberthAdam/jsonfile

The ``jsonfile`` Python module can be used to sync a JSON file
on a disk with the corresponding Python object instance.

Can be used to autosave JSON compatible Python data.
"""
__version__ = "0.0.9"

import collections.abc
import copy
import json
import pathlib
import tempfile


class JSONFileBase:

    @property
    def root(self):
        return self._root

    def _get_adapter_or_value(self, data):
        if isinstance(data, collections.abc.Mapping):
            return JSONFileObject(self._root, data)
        elif isinstance(data, str):
            return data
        elif isinstance(data, collections.abc.Sequence):
            return JSONFileArray(self._root, data)
        else:
            return data

    @staticmethod
    def _value_norm(value):
        if isinstance(value, (
                collections.abc.Mapping,
                JSONFileObject,
        )):
            return {str(k): v for k, v in value.items()}
            # JSON obejct keys must be strings
        elif isinstance(value, str):
            return value
        elif isinstance(value, (
                collections.abc.Sequence,
                JSONFileArray,
        )):
            return list(value)
        else:
            return value


class JSONFile(JSONFileBase):

    def __init__(self, filepath, *,
                 data=...,
                 default_data=...,
                 autosave=False,
                 dump_kwargs=None,
                 load_kwargs=None,
                 ):
        self._root = self
        self.changed = None
        self.default_data = default_data
        self._data = (default_data if data is ... else data)
        self.autosave = autosave
        self.dump_kwargs = dump_kwargs or dict(
            skipkeys=False,
            ensure_ascii=True,
            check_circular=True,
            allow_nan=True,
            cls=None,
            indent=None,
            separators=None,
            default=None,
            sort_keys=False,
        )
        self.load_kwargs = load_kwargs or dict(
            cls=None,
            object_hook=None,
            parse_float=None,
            parse_int=None,
            parse_constant=None,
            object_pairs_hook=None,
        )
        self.filepath = filepath
        if self._data == self.default_data:
            self.reload()
        elif autosave:
            self.save()

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.filepath}')"

    def __str__(self):
        return self.__repr__()

    @property
    def autosave(self):
        return bool(self._autosave)  # ensure boolean

    @autosave.setter
    def autosave(self, value):
        self._autosave = bool(value)  # ensure boolean

    @property
    def data(self):
        return self._get_adapter_or_value(self._data)

    @data.setter
    def data(self, value):
        if value is ...:
            raise ValueError("Ellipsis is forbidden, call delete()")
        old_data = copy.copy(self._data)
        self._data = self._value_norm(value)
        self.may_changed(self, old_data)

    @property
    def filepath(self):
        return self._filepath

    @filepath.setter
    def filepath(self, value):
        self._filepath = pathlib.Path(value)  # ensure Path instance

    def delete(self):
        # could have been @data.deleter but that would be obscure
        old_data = copy.copy(self._data)
        self._data = self.default_data
        self.may_changed(self, old_data)
        try:
            self.filepath.unlink()
        except FileNotFoundError:
            pass  # delete can happen previously

    def may_changed(self, inst, old_data):
        if inst._data != old_data:
            self.changed = True
            self.on_change()

    def on_change(self):
        # print("on change", self, self.autosave)
        if self.autosave:
            self.save()

    def reload(self):
        p = self.filepath
        if (
                p.is_file()  # filepath_exists
                and p.stat().st_size  # filepath_not_empty
        ):
            with p.open(encoding="utf8") as f:
                self._data = json.load(f, **self.load_kwargs)
                self.changed = False
        else:
            self._data = copy.deepcopy(self.default_data)
            self.changed = None

    def save(self, ensure_parents=True):
        p = self.filepath
        if self._data == self.default_data and p.is_file():
            p.unlink()
        else:
            s = json.dumps(self._data, **self.dump_kwargs)
            if ensure_parents:  # ensure parent directories
                p.parent.mkdir(parents=True, exist_ok=True)
            with tempfile.NamedTemporaryFile(
                    'w',
                    dir=p.parent,
                    delete=False,
                    encoding="utf8",
            ) as tf:
                tf.write(s)
                tp = p.parent / tf.name
            tp.replace(p)

from collections.abc import Mapping
class JSONFileContainer(JSONFileBase, Mapping):

    def __init__(self, root, data):
        self._root = root
        self._data = data

    def __contains__(self, key):  # has to be explicit
        return self._data.__contains__(key)

    def __delitem__(self, key):  # has to be explicit
        m = self._change_method("__delitem__")
        return m(key)

    def __eq__(self, other):  # has to be explicit
        return self._data.__eq__(other)

    def __dir__(self):
        return self._data.__dir__()

    @property
    def __doc__(self):
        return self._data.__doc__

    def __iter__(self):  # has to be explicit
        return self._data.__iter__()

    def __getattr__(self, name):
        return getattr(self._data, name)

    def __getattribute__(self, name):
        Ns = object.__getattribute__(self, "_change_method_names")
        if name in Ns:
            m = object.__getattribute__(self, "_change_method")
            return m(name)
        return object.__getattribute__(self, name)

    def __getitem__(self, key):
        data = self._data[key]
        return self._get_adapter_or_value(data)

    def __len__(self):
        return self._data.__len__()

    def __ne__(self, other):  # has to be explicit
        return self._data.__ne__(other)

    def __repr__(self):
        return self._data.__repr__()

    def __setitem__(self, key, value):  # has to be explicit
        m = self._change_method("__setitem__")
        return m(key, self._value_norm(value))

    def _change_method(self, name):
        def wrapped_method(*args, **kwargs):
            old_data = copy.copy(self._data)
            _data = object.__getattribute__(self, "_data")
            m = getattr(_data, name)
            try:
                r = m(*args, **kwargs)
                self._root.may_changed(self, old_data)
            except:
                self._data = old_data
                raise
            return r

        return wrapped_method


class JSONFileArray(JSONFileContainer, list):
    _change_method_names = (
        "__delitem__",
        "__iadd__",
        "__imul__",
        "__setitem__",
        "append",
        "clear",
        "extend",
        "insert",
        "pop",
        "remove",
        "reverse",
        "sort",
    )


class JSONFileObject(JSONFileContainer, dict):
    _change_method_names = (
        "__delitem__",
        "__setitem__",
        "clear",
        "pop",
        "popitem",
        # "setdefault",  # handled exceptionally
        "update",
    )

    def __setitem__(self, key, value):  # has to be explicit
        key = str(key)  # JSON obejct keys must be strings
        return super().__setitem__(key, value)

    def get(self, key, default=None):
        # avoid return of list/dict objects
        try:
            return self[key]
        except KeyError:
            return default

    def items(self):
        for k in self:
            yield k, self[k]

    def setdefault(self, key, default):
        try:
            return self[key]
        except KeyError:
            self[key] = default
        return self[key]

    def values(self):
        for k in self:
            yield self[k]

    def asdict(self):
        return self._data

jsonfile = JSONFile



#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import List, Dict, Optional

import PySimpleGUI as sg

from pep_tk.core.job import TaskStatus, TaskKey
from pep_tk.core.scheduler import SchedulerEventManager
from pep_tk.psg.events.data import ProgressGUIEventData


class GUIManager(SchedulerEventManager):
    def __init__(self, window: sg.Window, tabs: List['TaskTab']):
        super().__init__()
        self._window = window
        self._task_tabs: Dict[TaskKey, 'TaskTab'] = {t.task_key: t for t in tabs}
        self.stdout = {}
        self.stderr = {}

    def task_event_key(self, task_key: TaskKey):
        return self._task_tabs[task_key].task_progress_update_key

    def _initialize_task(self, task_key: TaskKey, count: int, max_count: int, status: TaskStatus):
        evt_data = ProgressGUIEventData(task_status=self.task_status[task_key],
                                        progress_count=count,
                                        max_count=max_count,
                                        elapsed_time=self.elapsed_time(task_key),
                                        completed_on_load=self.task_status[task_key] == TaskStatus.SUCCESS,
                                        output_log=self.pop_stdout(task_key, min_lines_to_pop=0))

        gui_task_event_key = self.task_event_key(task_key)
        self._window.write_event_value(gui_task_event_key, evt_data)
        print('task_initialized')

    def _check_cancelled(self, task_key: TaskKey):
        return self._task_tabs[task_key].is_cancelled

    def _start_task(self, task_key: TaskKey):
        print('task_started')

    def _end_task(self, task_key: TaskKey, status: TaskStatus):
        gui_task_event_key = self.task_event_key(task_key)
        evt_data = ProgressGUIEventData(task_status=status,
                                        progress_count=self.task_count[task_key],
                                        max_count=self.task_max_count[task_key],
                                        elapsed_time=self.elapsed_time(task_key),
                                        output_log=self.pop_stdout(task_key, min_lines_to_pop=0))
        self._window.write_event_value(gui_task_event_key, evt_data)
        # delete outputs from memory when task is finished
        if task_key in self.stdout:
            del self.stdout[task_key]
        if task_key in self.stderr:
            del self.stderr[task_key]
        print('task_finished')

    def _update_task_progress(self, task_key: TaskKey, current_count: int, max_count: int):
        evt_data = ProgressGUIEventData(task_status=self.task_status[task_key],
                                        progress_count=self.task_count[task_key],
                                        max_count=self.task_max_count[task_key],
                                        elapsed_time=self.elapsed_time(task_key),
                                        output_log=self.pop_stdout(task_key, min_lines_to_pop=5))

        gui_task_event_key = self.task_event_key(task_key)
        self._window.write_event_value(gui_task_event_key, evt_data)
        print('task_update_progress')

    def pop_stdout(self, task_key, min_lines_to_pop=50) -> Optional[str]:
        out = self.stdout.get(task_key)
        if not out:
            return None
        if len(out) > min_lines_to_pop:
            self.stdout[task_key] = []
            return ''.join(out)
        return None

    def _update_task_stdout(self, task_key: TaskKey, line: str):
        if task_key not in self.stdout:
            self.stdout[task_key] = []
        self.stdout[task_key].append(line)

    def _update_task_stderr(self, task_key: TaskKey, line: str):
        if task_key not in self.stderr:
            self.stderr[task_key] = ""
        self.stderr[task_key] += line
        # print('task_update_stderr(%s): %s' % (task_key, line))

    def _update_task_output_files(self, task_key: TaskKey, output_files: List[str]):
        gui_task_event_key = self.task_event_key(task_key)
        evt_data = ProgressGUIEventData(task_status=self.task_status[task_key],
                                        progress_count=self.task_count[task_key],
                                        max_count=self.task_max_count[task_key],
                                        elapsed_time=self.elapsed_time(task_key),
                                        output_files=output_files)
        self._window.write_event_value(gui_task_event_key, evt_data)

        print('_update_task_output_files')



#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import re
from typing import Dict

from pep_tk.core.configuration import PipelineConfig
from datetime import datetime

# Unfortunately the kwiver runner pipe-config will not inject environemnt variables
# and the -s flag creates new blocks sometimes instead of injecting setting in-place
# so here we compile a pipeline by basically replacing all $ENV with the intended value
# and also replacing relativepath attributes with their absolute path.
def compile_pipeline(pipeline: PipelineConfig, env: Dict) -> str:
    with open(pipeline.path, 'r') as f:
        pipeline_content = f.read()

    # Inject $ENV{} in pipeline with the intended value
    pipeline_required_env = list(set(re.findall(r"\$ENV{(.*)}", pipeline_content, re.M)))
    # for pipeline_required in pipeline_required_env:
    #     assert (pipeline_required in env)  # TODO: handle

    for k, v in env.items():
        search_str = '$ENV{%s}' % k
        pipeline_content = pipeline_content.replace(search_str, str(v))

    # # Custom pep_tk macro injection
    # pipeline_content = pipeline_content.replace('[DATASET]', dataset.name)  # [DATASET] macro

    # replace relativepath attributes with their absolute path
    relative_paths = list(set(re.findall(r"relativepath.*=\s*(.*)$", pipeline_content, re.M)))
    absolute_paths = {}
    for path in relative_paths:
        absolute_paths[path] = os.path.abspath(os.path.normpath(os.path.join(pipeline.directory, path)))

    for prev, new in absolute_paths.items():
        pipeline_content = pipeline_content.replace(prev, new)

    pipeline_content = re.sub(r"relativepath\s*", '', pipeline_content)
    return pipeline_content


def compile_output_filenames(output_filenames: Dict[str, str], path='', t=None) -> Dict[str, str]:
    out = {}
    t = t if t else datetime.now()
    timestr = t.strftime("%Y%m%d-%H%M%S")
    for k,v in output_filenames.items():
        new_v = v.replace('[TIMESTAMP]', timestr)
        new_v = os.path.join(path, new_v)
        out[k] = os.path.normpath(new_v)
    return out


#!/usr/bin/env python

#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

from distutils.core import setup
import sys
sys.path.append('src')
from setuptools import find_packages
import os
import unittest

# Parse Version information
version_txt_fp = os.path.normpath('src/pep_tk/VERSION.txt')
with open(version_txt_fp, 'r') as file:
    _v_ = file.read().replace('\n', '')
if len(_v_.split('.')) == 3:
    _v_ = _v_
else:
    print(f'Invalid VERSION.cfg - could not parse {_v_}')


# Parse requirements.txt to list
with open('requirements.txt') as f:
    install_reqs = f.read().splitlines()
install_reqs = list(filter(None, install_reqs))

# Parse README.md
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        path = os.sep.join(path.split(os.sep)[2:])
        for filename in filenames:
            paths.append(os.path.join(path, filename))
    return paths


extra_files = package_files('src/pep_tk/conf') + package_files('src/pep_tk/lib') + package_files('src/pep_tk')

# define function for python setup.py test
def test_suit():
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover('tests', pattern='test_*.py')
    return test_suite


setup(
        name='pep_tk',
        version=_v_,
        description='Polar Ecosystems Program Toolkit',
        long_description=long_description,
        long_description_content_type="text/markdown",
        author='Yuval Boss',
        author_email='yuval@uw.edu',
        url='https://github.com/readicculus/pep_gui',
        install_requires=install_reqs,
        packages=find_packages('src'),
        package_dir={'pep_tk': 'src/pep_tk'},
        package_data={"pep_tk": extra_files},
        include_package_data=True,
        entry_points={
          'console_scripts': ['pep_gui=pep_tk:launch.main'],
        },
        python_requires='>=3.8',
        test_suite="setup.test_suit",
)



#      This file is part of the PEP GUI detection pipeline batch running tool
#      Copyright (C) 2021 Yuval Boss yuval@uw.edu
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List

import PySimpleGUI as sg

from pep_tk.core.parser import ManifestParser
from pep_tk.psg.fonts import Fonts
from pep_tk.psg.layouts import LayoutSection, help_icon

class DatasetSelectionLayout(LayoutSection):
    def __init__(self, dataset_manifest: ManifestParser, event_key='_dataset_tab_', selected_datasets=None):
        self.dataset_manifest: ManifestParser = dataset_manifest
        self.datasets: List[str] = self.dataset_manifest.list_dataset_keys()
        self.selected_datasets = selected_datasets or []
        self.event_key = event_key
        self.right_button_key = '%s>' % self.event_key
        self.left_button_key = '%s<' % self.event_key
        self.selected_datasets_title_key = '%s_selected_datasets_title'

    def get_selected_datasets(self):
        return [self.dataset_manifest[ds_name] for ds_name in self.selected_datasets]

    def get_layout(self):
        dataset_column_size = (40, max(10 , min(15, len(self.datasets)))) # show at minimum 10 rows and at most 25 rows.
        layout = [[sg.T('Select a dataset or multiple datasets below.\n'
                        'Select multiple by clicking on multiple datasets then pressing the \'>\' button', font=Fonts.description),
                        help_icon('Use ctrl+click to select multiple, shift-click to select many.')],
                  [
                      sg.Column([
                          [sg.Text('Datasets', font=Fonts.tab_text)],
                                 [sg.Listbox(key='dataset_options',
                                             values=self.datasets,
                                             size=dataset_column_size,
                                             select_mode=sg.SELECT_MODE_EXTENDED,
                                             bind_return_key=True,
                                             font=Fonts.description,
                                             background_color='white')]
                      ]),
                      sg.Column([[sg.Text('\n')], [sg.Button('>', key=self.right_button_key)], [sg.Button('<', key=self.left_button_key)]]),
                      sg.Column([
                          [sg.Text('Selected Datasets %d' % len(self.selected_datasets),
                                   key=self.selected_datasets_title_key,
                                   size=(len('selected datasets x') + 5, 1),
                                   font=Fonts.tab_text)],
                          [sg.Listbox(key='selected_datasets',
                                      values=self.selected_datasets,
                                      size=dataset_column_size,
                                      select_mode=sg.SELECT_MODE_EXTENDED,
                                      bind_return_key=True,
                                      font=Fonts.description,
                                      background_color='white')]
                      ])
                  ],
                  [sg.Text('Filter:', font=Fonts.tab_text),
                   sg.InputText(key='datasets_filter', enable_events=True, size=(30, 1), background_color='white'),
                   sg.Button('clear filter', key='clear_filter', font=Fonts.tab_text)],
                  [sg.Text('', key='warning', text_color='red', size=(50, 1), font=Fonts.tab_text)]]
        return layout

    def handle(self, window, event, values):
        need_update = False
        if event == 'datasets_filter':
            exp = values['datasets_filter']
            if exp == '':
                self.datasets = self.dataset_manifest.list_dataset_keys()
            else:
                self.datasets = self.dataset_manifest.list_dataset_keys_txt(values['datasets_filter'] + '*')
            need_update = True
        elif event == 'clear_filter':
            window['datasets_filter'](value='')
            window.write_event_value('datasets_filter', '')
        elif event == self.right_button_key:
            selection = values['dataset_options']
            for s in selection:
                self.selected_datasets.append(s)

            need_update = True

        elif event == self.left_button_key:
            selected_datasets = values['selected_datasets']
            for s in selected_datasets:
                self.selected_datasets.remove(s)
            need_update = True

        if need_update:
            unselected_values = []
            for v in self.datasets:
                if not v in self.selected_datasets:
                    unselected_values.append(v)
            unselected_values = sorted(unselected_values)
            datasets_selected = sorted(self.selected_datasets)
            window['dataset_options'].Update(values=unselected_values)
            window['selected_datasets'].Update(values=datasets_selected)
            window[self.selected_datasets_title_key].Update(
                value='Selected Datasets (%d)' % len(self.selected_datasets))

    @property
    def layout_name(self):
        return 'Dataset Selection'
